Package: utf8
Title: Unicode Text Processing
Version: 1.2.5
Authors@R: 
    c(person(given = c("Patrick", "O."),
             family = "Perry",
             role = c("aut", "cph")),
      person(given = "Kirill",
             family = "M\u00fcller",
             role = "cre",
             email = "kirill@cynkra.com"),
      person(given = "Unicode, Inc.",
             role = c("cph", "dtc"),
             comment = "Unicode Character Database"))
Description: Process and print 'UTF-8' encoded international
    text (Unicode). Input, validate, normalize, encode, format, and
    display.
License: Apache License (== 2.0) | file LICENSE
URL: https://ptrckprry.com/r-utf8/, https://github.com/patperry/r-utf8
BugReports: https://github.com/patperry/r-utf8/issues
Depends: R (>= 2.10)
Suggests: cli, covr, knitr, rlang, rmarkdown, testthat (>= 3.0.0),
        withr
VignetteBuilder: knitr, rmarkdown
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.2.9000
NeedsCompilation: yes
Packaged: 2025-05-01 07:56:23 UTC; kirill
Author: Patrick O. Perry [aut, cph],
  Kirill Müller [cre],
  Unicode, Inc. [cph, dtc] (Unicode Character Database)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2025-05-01 08:20:02 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4; 2026-07-18 03:02:29 UTC; unix

Package: utf8
Title: Unicode Text Processing
Version: 1.2.6
Authors@R: 
    c(person(given = c("Patrick", "O."),
             family = "Perry",
             role = c("aut", "cph")),
      person(given = "Kirill",
             family = "M\u00fcller",
             role = "cre",
             email = "kirill@cynkra.com",
             comment = c(ORCID = "0000-0002-1416-3412")),
      person(given = "Unicode, Inc.",
             role = c("cph", "dtc"),
             comment = "Unicode Character Database"))
Description: Process and print 'UTF-8' encoded international
    text (Unicode). Input, validate, normalize, encode, format, and
    display.
License: Apache License (== 2.0) | file LICENSE
URL: https://krlmlr.github.io/utf8/, https://github.com/krlmlr/utf8
BugReports: https://github.com/krlmlr/utf8/issues
Depends: R (>= 2.10)
Suggests: cli, covr, knitr, rlang, rmarkdown, testthat (>= 3.0.0),
        withr
VignetteBuilder: knitr, rmarkdown
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.2.9000
NeedsCompilation: yes
Packaged: 2025-06-08 18:28:11 UTC; kirill
Author: Patrick O. Perry [aut, cph],
  Kirill Müller [cre] (ORCID: <https://orcid.org/0000-0002-1416-3412>),
  Unicode, Inc. [cph, dtc] (Unicode Character Database)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2025-06-08 19:00:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 08:46:38 UTC; unix
Archs: utf8.so.dSYM

# Generated by roxygen2: do not edit by hand

export(as_utf8)
export(output_ansi)
export(output_utf8)
export(utf8_encode)
export(utf8_format)
export(utf8_normalize)
export(utf8_print)
export(utf8_valid)
export(utf8_width)
useDynLib(utf8, .registration = TRUE)

<!-- README.md and index.md are generated from README.Rmd. Please edit that file. -->

# utf8

<!-- badges: start -->

[![rcc](https://github.com/patperry/r-utf8/workflows/rcc/badge.svg)](https://github.com/krlmlr/utf8/actions)
[![Coverage
Status](https://codecov.io/github/patperry/r-utf8/coverage.svg?branch=main "Code Coverage")](https://app.codecov.io/github/patperry/r-utf8?branch=main "Code Coverage")
[![CRAN
Status](https://www.r-pkg.org/badges/version/utf8 "CRAN Page")](https://cran.r-project.org/package=utf8 "CRAN Page")
[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg "Apache License, Version 2.0")](https://www.apache.org/licenses/LICENSE-2.0.html "Apache License, Version 2.0")
[![CRAN RStudio Mirror
Downloads](https://cranlogs.r-pkg.org/badges/utf8 "CRAN Downloads")](https://cran.r-project.org/package=utf8 "CRAN Page")
<!-- badges: end -->

*utf8* is an R package for manipulating and printing UTF-8 text that
fixes multiple bugs in R’s UTF-8 handling.

## Installation

### Stable version

*utf8* is [available on
CRAN](https://cran.r-project.org/package=utf8 "CRAN Page"). To install
the latest released version, run the following command in R:

``` r
install.packages("utf8")
```

### Development version

To install the latest development version, run the following:

``` r
devtools::install_github("patperry/r-utf8")
```

## Usage

``` r
library(utf8)
```

### Validate character data and convert to UTF-8

Use `as_utf8()` to validate input text and convert to UTF-8 encoding.
The function alerts you if the input text has the wrong declared
encoding:

``` r
# second entry is encoded in latin-1, but declared as UTF-8
x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
Encoding(x) <- c("UTF-8", "UTF-8", "bytes")
as_utf8(x) # fails
#> Error in as_utf8(x): entry 2 has wrong Encoding; marked as "UTF-8" but leading byte 0xE7 followed by invalid continuation byte (0xdeadbeef) at position 4

# mark the correct encoding
Encoding(x[2]) <- "latin1"
as_utf8(x) # succeeds
#> [1] "façile" "façile" "façile"
```

### Normalize data

Use `utf8_normalize()` to convert to Unicode composed normal form (NFC).
Optionally apply compatibility maps for NFKC normal form or case-fold.

``` r
# three ways to encode an angstrom character
(angstrom <- c("\u00c5", "\u0041\u030a", "\u212b"))
#> [1] "Å" "Å" "Å"
utf8_normalize(angstrom) == "\u00c5"
#> [1] TRUE TRUE TRUE

# perform full Unicode case-folding
utf8_normalize("Größe", map_case = TRUE)
#> [1] "grösse"

# apply compatibility maps to NFKC normal form
# (example from https://twitter.com/aprilarcus/status/367557195186970624)
utf8_normalize("𝖸𝗈 𝐔𝐧𝐢𝐜𝐨𝐝𝐞 𝗅 𝗁𝖾𝗋𝖽 𝕌 𝗅𝗂𝗄𝖾 𝑡𝑦𝑝𝑒𝑓𝑎𝑐𝑒𝑠 𝗌𝗈 𝗐𝖾 𝗉𝗎𝗍 𝗌𝗈𝗆𝖾 𝚌𝚘𝚍𝚎𝚙𝚘𝚒𝚗𝚝𝚜 𝗂𝗇 𝗒𝗈𝗎𝗋 𝔖𝔲𝔭𝔭𝔩𝔢𝔪𝔢𝔫𝔱𝔞𝔯𝔶 𝔚𝔲𝔩𝔱𝔦𝔩𝔦𝔫𝔤𝔳𝔞𝔩 𝔓𝔩𝔞𝔫𝔢 𝗌𝗈 𝗒𝗈𝗎 𝖼𝖺𝗇 𝓮𝓷𝓬𝓸𝓭𝓮 𝕗𝕠𝕟𝕥𝕤 𝗂𝗇 𝗒𝗈𝗎𝗋 𝒇𝒐𝒏𝒕𝒔.",
               map_compat = TRUE)
#> [1] "Yo Unicode l herd U like typefaces so we put some codepoints in your Supplementary Wultilingval Plane so you can encode fonts in your fonts."
```

### Print emoji

On some platforms (including MacOS), the R implementation of `print()`
uses an outdated version of the Unicode standard to determine which
characters are printable. Use `utf8_print()` for an updated print
function:

``` r
print(intToUtf8(0xdeadbeefF600 + 0:79)) # with default R print function
#> [1] "😀😁😂😃😄😅😆😇😈😉😊😋😌😍😎😏😐😑😒😓😔😕😖😗😘😙😚😛😜😝😞😟😠😡😢😣😤😥😦😧😨😩😪😫😬😭😮😯😰😱😲😳😴😵😶😷😸😹😺😻😼😽😾😿🙀🙁🙂🙃🙄🙅🙆🙇🙈🙉🙊🙋🙌🙍🙎🙏"

utf8_print(intToUtf8(0xdeadbeefF600 + 0:79)) # with utf8_print, truncates line
#> [1] "😀​😁​😂​😃​😄​😅​😆​😇​😈​😉​😊​😋​😌​😍​😎​😏​😐​😑​😒​😓​😔​😕​😖​😗​😘​😙​😚​😛​😜​😝​😞​😟​😠​😡​😢​😣​…"

utf8_print(intToUtf8(0xdeadbeefF600 + 0:79), chars = 1000) # higher character limit
#> [1] "😀​😁​😂​😃​😄​😅​😆​😇​😈​😉​😊​😋​😌​😍​😎​😏​😐​😑​😒​😓​😔​😕​😖​😗​😘​😙​😚​😛​😜​😝​😞​😟​😠​😡​😢​😣​😤​😥​😦​😧​😨​😩​😪​😫​😬​😭​😮​😯​😰​😱​😲​😳​😴​😵​😶​😷​😸​😹​😺​😻​😼​😽​😾​😿​🙀​🙁​🙂​🙃​🙄​🙅​🙆​🙇​🙈​🙉​🙊​🙋​🙌​🙍​🙎​🙏​"
```

## Citation

Cite *utf8* with the following BibTeX entry:

    @Manual{,
      title = {utf8: Unicode Text Processing},
      author = {Patrick O. Perry},
      note = {R package version 1.2.4.9900, https://github.com/patperry/r-utf8},
      url = {https://ptrckprry.com/r-utf8/},
    }

## Contributing

The project maintainer welcomes contributions in the form of feature
requests, bug reports, comments, unit tests, vignettes, or other code.
If you’d like to contribute, either

- fork the repository and submit a pull request

- [file an issue](https://github.com/krlmlr/utf8/issues "Issues");

- or contact the maintainer via e-mail.

This project is released with a [Contributor Code of
Conduct](https://github.com/krlmlr/utf8/blob/main/CONDUCT.md "Contributor Code of Conduct"),
and if you choose to contribute, you must adhere to its terms.

utf8lite (C Library)
====================

[![Build Status](https://api.travis-ci.org/patperry/utf8lite.svg?branch=master)](https://travis-ci.org/patperry/utf8lite)
[![Coverage Status](https://codecov.io/github/patperry/utf8lite/coverage.svg?branch=master)](https://codecov.io/github/patperry/utf8lite?branch=master)

Lightweight UTF-8 processing.

local_ctype <- function(mode = c("C", "UTF-8"), env = parent.frame()) {
  mode <- match.arg(mode)

  if (mode == "UTF-8") {
    sysname <- Sys.info()[["sysname"]]
    if (sysname == "Windows") {
      ctype <- "English_United States.1252"
    } else if (sysname == "Darwin") {
      ctype <- "UTF-8"
    } else {
      ctype <- "en_US.utf8"
    }
  } else {
    ctype <- "C"
  }

  ctype0 <- Sys.getlocale("LC_CTYPE")
  suppressWarnings({
    withr::local_locale(LC_CTYPE = ctype, .local_envir = env)
  })
  if (Sys.getlocale("LC_CTYPE") != ctype) {
    skip(paste0("Cannot change locale to '", ctype, "'"))
  }
  if (mode == "UTF-8" && !output_utf8()) {
    skip("Cannot change to UTF-8 output")
  }

  ctype0
}

with_ctype <- function(mode = c("C", "UTF-8"), code) {
  local_ctype(mode)
  force(code)
}

test_that("'format' can handle short text", {
  raw <- c(NA, "", "a", "foo", "short text")

  expect_equal(
    utf8_format(raw, justify = "none", na.print = "NA"),
    format(raw, justify = "none")
  )

  expect_equal(
    utf8_format(raw, justify = "left", na.print = "NA"),
    format(raw, justify = "left")
  )

  expect_equal(
    utf8_format(raw, justify = "centre", na.print = "NA"),
    format(raw, justify = "centre")
  )

  expect_equal(
    utf8_format(raw, justify = "right", na.print = "NA"),
    format(raw, justify = "right")
  )
})


test_that(paste0("'format' can handle long text in Unicode locale"), {
  if (l10n_info()$`UTF-8`) {
    local_utf8()
    expect_true(cli::is_utf8_output())
    mapper <- identity
    variant <- "unicode"
  } else {
    mapper <- function(x) lapply(x, charToRaw)
    variant <- "ansi"
  }

  skip_if_not_installed("testthat", "3.1.1")

  expect_snapshot(variant = variant, {
    raw <- c(
      NA, "", "a", "ab", "foo", "food", "short text",
      "\u6027", "\u6027\u6027", "\u6027?"
    )
    Encoding(raw) <- "UTF-8"

    body(mapper)

    mapper(utf8_format(raw, chars = 2, justify = "none", na.print = "NA"))
    mapper(utf8_format(raw, chars = 2, justify = "left", na.print = "NA"))
    mapper(utf8_format(raw, chars = 2, justify = "centre", na.print = "NA"))
    mapper(utf8_format(raw, chars = 2, justify = "right", na.print = "NA"))
  })
})

test_that("'format' can handle long text in UTF-8 locale, part 2", {
  raw <- c(
    NA, "", "a", "\n", "ab", "foo", "food", "short text",
    "\u6027", "\u6027\u6027", "\u6027?"
  )
  short <- c(
    NA, "", "a", "\u2026", "a\u2026", "f\u2026", "f\u2026",
    "s\u2026", "\u2026", "\u2026", "\u2026"
  )
  rshort <- c(
    NA, "", "a", "\u2026", "\u2026b", "\u2026o", "\u2026d",
    "\u2026t", "\u2026", "\u2026", "\u2026?"
  )

  local_ctype("UTF-8")

  expect_equal(
    utf8_format(raw,
      chars = 1, justify = "none",
      na.encode = FALSE
    ),
    format(short, justify = "none", na.encode = FALSE)
  )

  expect_equal(
    utf8_format(raw, chars = 1, justify = "left", na.encode = FALSE),
    format(short, justify = "left", na.encode = FALSE)
  )

  expect_equal(
    utf8_format(raw, chars = 1, justify = "centre", na.encode = FALSE),
    format(short, justify = "centre", na.encode = FALSE)
  )

  expect_equal(
    utf8_format(raw,
      chars = 1, justify = "right",
      na.encode = FALSE
    ),
    format(rshort, justify = "right", na.encode = FALSE)
  )
})


test_that("'format' can handle long text in C locale", {
  #            6         7          8           9            10
  raw <- c(
    "\u6027", "?\u6027", "\n\u6027", "?\n\u6027", "\u0001\u6027",
    "\u6027?", "\u6027\n", "\u6027?\n", "\u6027\u0001"
  )
  short <- c(
    "\\u6027", "?\\u6027", "\\n\\u6027", "?\\n...", "\\u0001...",
    "\\u6027?", "\\u6027\\n", "\\u6027?...", "\\u6027..."
  )
  rshort <- c(
    "\\u6027", "?\\u6027", "\\n\\u6027", "...\\n\\u6027",
    "...\\u6027", "\\u6027?", "\\u6027\\n", "...?\\n",
    "...\\u0001"
  )

  skip_on_os("windows")

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    short_format <- utf8_encode(utf8_format(raw, chars = 8, justify = "none"))
    left <- utf8_encode(utf8_format(raw, chars = 8, justify = "left"))
    centre <- utf8_encode(utf8_format(raw, chars = 8, justify = "centre"))
    right <- utf8_encode(utf8_format(raw, chars = 8, justify = "right"))
  })

  expect_equal(short_format, format(short, justify = "none"))

  expect_equal(sub("\\s+$", "", left), short)
  expect_equal(as.numeric(nchar(left)), rep(10, length(raw)))

  expect_equal(sub("^\\s+", "", sub("\\s+$", "", centre)), short)
  expect_equal(as.numeric(nchar(centre)), rep(10, length(raw)))

  expect_equal(sub("^\\s+", "", right), rshort)
  expect_equal(as.numeric(nchar(right)), rep(11, length(raw)))
})


test_that("'format' can handle high code points in C locale", {
  raw <- c(intToUtf8(0x00010000), intToUtf8(0x0010ffff))

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    left <- utf8_format(raw, justify = "left")
    right <- utf8_format(raw, justify = "right")
  })

  expect_equal(left, raw)
  expect_equal(right, raw)
})


test_that("'format' can handle high code points in Unicode locale", {
  local_ctype("UTF-8")
  skip_on_os("windows") # no Unicode above 0xFFFF on Windows

  raw <- c(intToUtf8(0x00010000), intToUtf8(0x010ffff))
  left <- c(paste0(intToUtf8(0x00010000), "         "), intToUtf8(0x010ffff))
  right <- c(paste0("         ", intToUtf8(0x00010000)), intToUtf8(0x010ffff))

  expect_equal(utf8_format(raw, justify = "left"), left)
  expect_equal(utf8_format(raw, justify = "right"), right)
})


test_that("'format' can handle ignorable code points", {
  raw <- "\u200B"

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    left <- utf8_format(raw, justify = "left")
    centre <- utf8_format(raw, justify = "centre")
    right <- utf8_format(raw, justify = "right")
  })

  expect_equal(as.character(left), raw)
  expect_equal(as.character(centre), raw)
  expect_equal(as.character(right), raw)

  local_ctype("UTF-8")

  expect_equal(utf8_format(raw, justify = "left"), raw)
  expect_equal(utf8_format(raw, justify = "centre"), raw)
  expect_equal(utf8_format(raw, justify = "right"), raw)
})

test_that("'format' can handle marks in C locale", {
  skip_on_os("windows")

  raw <- "\u1e0d\u0307"

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    left <- utf8_format(raw, chars = 6, justify = "left")
    centre <- utf8_format(raw, chars = 6, justify = "centre")
    right <- utf8_format(raw, chars = 6, justify = "right")
  })

  expect_equal(as.character(left), "...")
  expect_equal(as.character(centre), "...")
  expect_equal(as.character(right), "...")
})

test_that("'format' can handle marks", {
  raw <- "\u1e0d\u0307"

  local_ctype("UTF-8")

  expect_equal(utf8_format(raw, chars = 1, justify = "left"), raw)
  expect_equal(utf8_format(raw, chars = 1, justify = "centre"), raw)
  expect_equal(utf8_format(raw, chars = 1, justify = "right"), raw)
})


test_that("'format' can handle UTF-8 'Other' codes", {
  raw <- "\u2072" # unassigned

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    left <- utf8_format(raw, justify = "left")
    centre <- utf8_format(raw, justify = "centre")
    right <- utf8_format(raw, justify = "right")
  })

  expect_equal(as.character(left), raw)
  expect_equal(as.character(centre), raw)
  expect_equal(as.character(right), raw)

  local_ctype("UTF-8")

  expect_equal(utf8_format(raw, justify = "left"), raw)
  expect_equal(utf8_format(raw, justify = "centre"), raw)
  expect_equal(utf8_format(raw, justify = "right"), raw)
})


test_that("'format' can handle zero chars", {
  skip_on_os("windows")

  raw <- "foo"

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    left <- utf8_format(raw, chars = 0, justify = "left")
    centre <- utf8_format(raw, chars = 0, justify = "centre")
    right <- utf8_format(raw, chars = 0, justify = "right")
  })

  expect_equal(as.character(left), "...")
  expect_equal(as.character(centre), "...")
  expect_equal(as.character(right), "...")
})


test_that("'format' can handle NULL chars", {
  raw <- "foo"

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    left <- utf8_format(raw, chars = NULL, justify = "left")
    centre <- utf8_format(raw, chars = NULL, justify = "centre")
    right <- utf8_format(raw, chars = NULL, justify = "right")
  })

  expect_equal(as.character(left), "foo")
  expect_equal(as.character(centre), "foo")
  expect_equal(as.character(right), "foo")
})


test_that("'format' can skip NA", {
  expect_equal(
    as.character(utf8_format(NA_character_, na.encode = FALSE)),
    NA_character_
  )
})


test_that("'format' can set minimum width", {
  raw <- c("a", "ab", "abc")

  expect_equal(
    utf8_format(raw, justify = "none", width = 5),
    format(raw, justify = "none", width = 5)
  )
  expect_equal(
    utf8_format(raw, justify = "left", width = 5),
    format(raw, justify = "left", width = 5)
  )
  expect_equal(
    utf8_format(raw, justify = "centre", width = 5),
    format(raw, justify = "centre", width = 5)
  )
  expect_equal(
    utf8_format(raw, justify = "right", width = 5),
    format(raw, justify = "right", width = 5)
  )
})


test_that("'format' error for invalid justify", {
  expect_error(utf8_format("", justify = "wild"),
    paste(
      "'justify' must be one of the following:",
      paste(dQuote(c("left", "right", "centre", "none")),
        collapse = ", "
      )
    ),
    fixed = TRUE
  )
})


test_that("'format' error for invalid logicals", {
  expect_error(utf8_format("", trim = NA), "'trim' must be TRUE or FALSE",
    fixed = TRUE
  )
  expect_error(utf8_format("", na.encode = NA),
    "'na.encode' must be TRUE or FALSE",
    fixed = TRUE
  )
})


test_that("'format' error for invalid integers", {
  expect_error(utf8_format("", chars = "3"),
    "'chars' must be integer-valued",
    fixed = TRUE
  )
  expect_error(utf8_format("", width = "3"),
    "'width' must be integer-valued",
    fixed = TRUE
  )
})


test_that("'utf8_format' can handle invalid UTF-8", {
  input <- "\xff\xfe"
  Encoding(input) <- "UTF-8"
  x <- c("a", input)
  output <- input
  Encoding(output) <- "bytes"

  expect_equal(utf8_format(x, justify = "none"), c("a", output))
  expect_equal(utf8_format(x, justify = "left"), c("a       ", output))
  expect_equal(utf8_format(x, justify = "centre"), c("   a    ", output))
  expect_equal(utf8_format(x, justify = "right"), c("       a", output))
})


test_that("'utf8_format' can handle latin1 text", {
  x <- "fa\xE7ile"
  Encoding(x) <- "latin1"
  y <- iconv(x, "latin1", "UTF-8")
  expect_equal(utf8_format(x), y)
})


test_that("'utf8_format' can quote", {
  expect_equal(
    utf8_format(c("a", "abc"), quote = TRUE),
    c("a  ", "abc")
  )

  expect_equal(
    utf8_format(c("a", "abc"), quote = TRUE, justify = "centre"),
    c(" a ", "abc")
  )

  expect_equal(
    utf8_format(c("a", "abc"), quote = TRUE, justify = "right"),
    c("  a", "abc")
  )
})


test_that("'utf8_format' can handle quotes", {
  expect_equal(utf8_format('"'), '"')
  expect_equal(utf8_format('"', quote = TRUE), '"')
  expect_equal(utf8_format('"', justify = "right"), '"')
  expect_equal(utf8_format('"', justify = "right", quote = TRUE), '"')
})


test_that("'utf8_format' works on bytes", {
  x <- "fa\xC3\xA7ile"
  Encoding(x) <- "bytes"

  l <- utf8_format(x, justify = "left")
  expect_equal(Encoding(l), "bytes")

  r <- utf8_format(x, justify = "right")
  expect_equal(Encoding(r), "bytes")

  Encoding(x) <- Encoding(l) <- Encoding(r) <- "UTF-8"
  expect_equal(l, x)
  expect_equal(r, x)
})


test_that("'utf8_format' can right justify", {
  x <- " (3322 rows total)"
  expect_equal(
    utf8_format(x, width = 79, justify = "right"),
    format(x, width = 79, justify = "right")
  )
})


test_that("'utf8_format' use ... ellipsis for bytes", {
  local_ctype("UTF-8")

  x <- "fa\xC3\xA7ile"
  Encoding(x) <- "bytes"
  y <- "fa\xC3..."
  Encoding(y) <- "bytes"

  expect_equal(utf8_format(x, chars = 6), y)
})

test_that("'utf8_width' computes widths correctly", {
  local_ctype("UTF-8")

  expect_equal(
    utf8_width(c("hello", "\u200b", "\u22ee", "\u6027"),
      encode = FALSE
    ),
    c(5, 0, 1, 2)
  )
})


test_that("'utf8_width' computes widths for extended unicode correctly", {
  local_ctype("UTF-8")
  skip_on_os("windows") # no extended Unicode

  expect_equal(utf8_width(intToUtf8(0x1f642), encode = FALSE), 2)
})




test_that("'utf8_width' gives NA for non-ASCII in C locale", {
  x <- c("hello", "\u200b", "\u22ee", "\u6027", intToUtf8(0x1f642))

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    width <- utf8_width(x, encode = FALSE)
  })

  expect_equal(width, c(5, NA, NA, NA, NA))
})


test_that("'utf8_width' keeps names", {
  expect_equal(
    utf8_width(c(a = "hello", b = "you")),
    c(a = 5, b = 3)
  )
})


test_that("'utf8_width' gives NA for control characters", {
  expect_equal(
    utf8_width(c("\u0001", "\a", "new\nline"),
      encode = FALSE
    ),
    c(NA_integer_, NA_integer_, NA_integer_)
  )
})


test_that("'utf8_width' gives NA for invalid data", {
  x <- c("a", "b", "\xff", "abc\xfe")
  Encoding(x) <- "UTF-8"
  expect_equal(utf8_width(x, encode = FALSE), c(1, 1, NA, NA))
})


test_that("'utf8_width' gives width 1 for quotes", {
  expect_equal(utf8_width("\"", encode = TRUE), 1)
  expect_equal(utf8_width("\"", encode = FALSE), 1)
})


test_that("'utf8_width' gives correct with in C locale", {
  x <- intToUtf8(c(0x1F487, 0x200D, 0x2642, 0xFE0F))

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    width <- utf8_width(x)
    encoded <- utf8_encode(x)
  })

  expect_equal(width, nchar(encoded))
})

test_that("'utf8_encode' can encode an ASCII string", {
  expect_equal(utf8_encode("hello"), "hello")
})


test_that("'utf8_encode' can encode NULL or an NA string", {
  expect_equal(utf8_encode(NULL), NULL)
  expect_equal(utf8_encode(NA_character_), NA_character_)
})


test_that("'utf8_encode' preserves attributes", {
  x <- matrix(LETTERS, 2, 13)
  x[1, 4] <- "\xa4"
  Encoding(x) <- "latin1"
  dimnames(x) <- list(c("foo", "bar"), as.character(1:13))
  class(x) <- "my_class"

  local_ctype("UTF-8")

  expect_equal(utf8_encode(x), enc2utf8(x))
})


test_that("'utf8_encode' can encode basic Unicode", {
  x <- "\u200b"
  Encoding(x) <- "UTF-8"

  local_ctype("UTF-8")

  expect_equal(utf8_encode(x), x)
})


test_that("'utf8_encode' can encode extended Unicode", {
  skip_on_os("windows") # no extended Unicode

  x <- intToUtf8(0x0001f60d)
  Encoding(x) <- "UTF-8"

  local_ctype("UTF-8")

  expect_equal(utf8_encode(x), x)
})


test_that("'utf8_encode' can handle ASCII escapes", {
  x <- "\x01\a\b\f\n\r\t\v\x7f"
  expect_equal(utf8_encode(x), "\\u0001\\a\\b\\f\\n\\r\\t\\v\\u007f")
})

test_that("'utf8_encode' can handle invalid UTF-8", {
  x <- "\xfe"
  Encoding(x) <- "bytes"
  expect_equal(utf8_encode(x), "\\xfe")
})

test_that("'utf8_encode' can handle bytes", {
  x <- "\x01\a\b\f\n\r\t\v\x7f\x80\xff"
  Encoding(x) <- "bytes"
  expect_equal(
    utf8_encode(x),
    "\\x01\\a\\b\\f\\n\\r\\t\\v\\x7f\\x80\\xff"
  )
})

test_that("'utf8_encode' can handle latin-1", {
  x <- "her \xa320"
  Encoding(x) <- "latin1"

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    latin <- utf8_encode(x)
  })
  expect_equal(latin, "her \\u00a320")

  local_ctype("UTF-8")
  expect_equal(utf8_encode(x), "her \u00a320")
})


test_that("'utf8_encode' can handle bytes", {
  x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
  Encoding(x) <- c("UTF-8", "UTF-8", "bytes")

  local_ctype("UTF-8")

  y <- c("fa\u00e7ile", "fa\\xe7ile", "fa\\xc3\\xa7ile")
  Encoding(y) <- c("UTF-8", "UTF-8", "bytes")

  expect_equal(utf8_encode(x), y)
})


test_that("'utf8_encode escapes controls in UTF-8 text", {
  local_ctype("UTF-8")

  x <- "\n\u2026"
  Encoding(x) <- "UTF-8"
  y <- "\\n\u2026"
  Encoding(y) <- "UTF-8"
  expect_equal(utf8_encode(x), y)
})


test_that("'utf8_encode' can quote", {
  x <- c("abcde", "x", "123", "\"", "'")
  expect_equal(utf8_encode(x, quote = FALSE), encodeString(x, quote = ""))
  expect_equal(utf8_encode(x, quote = TRUE), encodeString(x, quote = '"'))
})


test_that("'utf8_encode' ignores justify if width = 0", {
  x <- c("abcde", "x", "123", "\"", "'")
  expect_equal(
    utf8_encode(x, width = 0, justify = "none"),
    encodeString(x, width = 0, justify = "none")
  )
  expect_equal(
    utf8_encode(x, width = 0, justify = "left"),
    encodeString(x, width = 0, justify = "left")
  )
  expect_equal(
    utf8_encode(x, width = 0, justify = "centre"),
    encodeString(x, width = 0, justify = "centre")
  )
  expect_equal(
    utf8_encode(x, width = 0, justify = "right"),
    encodeString(x, width = 0, justify = "right")
  )
})


test_that("'utf8_encode' can justify", {
  x <- c("abcde", "x", "123", "\"", "'")
  expect_equal(
    utf8_encode(x, width = NULL, justify = "none"),
    encodeString(x, width = NULL, justify = "none")
  )
  expect_equal(
    utf8_encode(x, width = NULL, justify = "left"),
    encodeString(x, width = NULL, justify = "left")
  )
  expect_equal(
    utf8_encode(x, width = NULL, justify = "centre"),
    encodeString(x, width = NULL, justify = "centre")
  )
  expect_equal(
    utf8_encode(x, width = NULL, justify = "right"),
    encodeString(x, width = NULL, justify = "right")
  )
})


test_that("'utf8_encode' can justify and quote", {
  expect_equal(
    utf8_encode(c("1", "10", "100"), width = NULL, quote = TRUE),
    encodeString(c("1", "10", "100"), width = NULL, quote = '"')
  )
})

test_that("'utf8_print' can print unicode", {
  skip_on_os("windows")
  local_ctype("UTF-8")

  x <- c(
    "\u0100\u0101\u0102\u0103\u0104\u0105",
    "\u0106\u0107\u0108\u0109\u010a\u010b"
  )

  expect_equal(
    capture_output(utf8_print(x)),
    paste(
      "[1] \"\u0100\u0101\u0102\u0103\u0104\u0105\"",
      "\"\u0106\u0107\u0108\u0109\u010a\u010b\""
    )
  )
})

test_that("'utf8_print' works with unnamed character vectors", {
  x <- as.character(1:100)

  expect_equal(
    capture_output(utf8_print(x)),
    capture_output(print(x))
  )

  expect_equal(
    capture_output(utf8_print(x[1:96])),
    capture_output(print(x[1:96]))
  )

  expect_equal(
    capture_output(utf8_print(x[1:7])),
    capture_output(print(x[1:7]))
  )
})


test_that("'utf8_print' works with named character vectors", {
  x <- as.character(10 + 1:26)
  names(x) <- letters

  # left align names
  xr <- x
  names(xr) <- format(names(x), aligh = "left", width = 4)
  actual <- strsplit(capture_output(utf8_print(x)), "\n")[[1]]
  expected <- strsplit(capture_output(print(xr)), "\n")[[1]]
  expect_equal(paste(actual, ""), expected)

  actual <- strsplit(capture_output(utf8_print(x[1:16])), "\n")[[1]]
  expected <- strsplit(capture_output(print(xr[1:16])), "\n")[[1]]
  expect_equal(paste(actual, ""), expected)

  actual <- strsplit(capture_output(utf8_print(x[1:4])), "\n")[[1]]
  expected <- strsplit(capture_output(print(xr[1:4])), "\n")[[1]]
  expect_equal(paste(actual, ""), expected)
})


test_that("'utf8_print' can use the 'max' argument for unnamed vectors", {
  x <- as.character(1:100)

  expect_equal(
    capture_output(utf8_print(x, max = 0), width = 80),
    " [ reached getOption(\"max.print\") -- omitted 100 entries ]"
  )

  expect_equal(
    capture_output(utf8_print(x, max = 100), width = 80),
    capture_output(utf8_print(x), width = 80)
  )

  lines <- strsplit(
    capture_output(utf8_print(x, max = 20), width = 80),
    "\n"
  )[[1]]
  expect_equal(length(lines), 3)
  expect_equal(
    lines[[3]],
    " [ reached getOption(\"max.print\") -- omitted 80 entries ]"
  )
})


test_that("'utf8_print' can use the 'max' argument for named vectors", {
  x <- as.character(1:260)
  names(x) <- rep(letters, 10)

  expect_equal(
    capture_output(utf8_print(x, max = 0), width = 80),
    " [ reached getOption(\"max.print\") -- omitted 260 entries ]"
  )

  expect_equal(
    capture_output(utf8_print(x, max = 260), width = 80),
    capture_output(utf8_print(x), width = 80)
  )

  lines <- strsplit(
    capture_output(utf8_print(x, max = 20), width = 80),
    "\n"
  )[[1]]
  expect_equal(length(lines), 5)
  expect_equal(
    lines[[5]],
    " [ reached getOption(\"max.print\") -- omitted 240 entries ]"
  )
})


test_that("'utf8_print' can print empty vectors", {
  expect_equal(capture_output(utf8_print(character())), "character(0)")
  expect_equal(capture_output(utf8_print(array(character(), 0))), "character(0)")
})


test_that("'utf8_print' can print matrices", {
  x1 <- matrix(letters, 13, 2)

  x2 <- matrix(letters, 13, 2)
  rownames(x2) <- LETTERS[1:13]

  x3 <- matrix(letters, 13, 2)
  colnames(x3) <- c("x", "y")

  x4 <- matrix(letters, 13, 2)
  rownames(x4) <- LETTERS[1:13]
  colnames(x4) <- c("x", "y")

  expect_equal(
    capture_output(utf8_print(x1)),
    capture_output(print(x1))
  )

  expect_equal(
    capture_output(utf8_print(x2)),
    capture_output(print(x2))
  )

  expect_equal(
    capture_output(utf8_print(x3)),
    capture_output(print(x3))
  )

  expect_equal(
    capture_output(utf8_print(x4)),
    capture_output(print(x4))
  )
})


test_that("'utf8_print' can print empty matrices", {
  x1 <- matrix(character(), 10, 0)
  x2 <- matrix(character(), 0, 10)
  x3 <- matrix(character(), 0, 0)

  expect_equal(
    paste0("     \n", capture_output(utf8_print(x1))),
    capture_output(print(x1))
  )

  expect_equal(
    paste0("     ", capture_output(utf8_print(x2))),
    capture_output(print(x2))
  )

  expect_equal(
    capture_output(utf8_print(x3)),
    capture_output(print(x3))
  )
})


test_that("'utf8_print' can print arrays", {
  x <- array(as.character(1:24), c(2, 3, 4, 5))

  expect_equal(
    capture_output(utf8_print(x)),
    capture_output(print(x))
  )

  x2 <- x
  dimnames(x2) <- list(
    letters[1:2], letters[3:5], letters[6:9],
    letters[10:14]
  )

  expect_equal(
    capture_output(utf8_print(x2)),
    capture_output(print(x2))
  )
})


test_that("'utf8_print' can print empty arrays", {
  expect_equal(
    capture_output(utf8_print(array(character(), c(2, 3, 0)))),
    "<2 x 3 x 0 array>"
  )

  expect_equal(
    capture_output(utf8_print(array(character(), c(2, 0, 3)))),
    "<2 x 0 x 3 array>"
  )

  expect_equal(
    capture_output(utf8_print(array(character(), c(0, 2, 3)))),
    "<0 x 2 x 3 array>"
  )
})


test_that("'utf8_print' can print quotes", {
  expect_equal(
    capture_output(utf8_print('"')),
    capture_output(print('"'))
  )

  expect_equal(
    capture_output(utf8_print('"', quote = FALSE)),
    capture_output(print('"', quote = FALSE))
  )
})


test_that("'utf8_print' can handle NA", {
  expect_equal(
    capture_output(utf8_print(NA_character_)),
    capture_output(print(NA_character_))
  )
  expect_equal(
    capture_output(utf8_print(NA_character_, quote = FALSE)),
    capture_output(print(NA_character_, quote = FALSE))
  )
})


test_that("'utf8_print' can handle NA names", {
  x <- matrix("hello", 1, 1, dimnames = list(NA, NA))
  expect_equal(
    capture_output(utf8_print(x)),
    capture_output(print(x))
  )
  expect_equal(
    capture_output(utf8_print(x, na.print = "foo")),
    capture_output(print(x, na.print = "foo"))
  )
})


test_that("'utf8_print' can right justify", {
  x <- matrix(c("a", "ab", "abc"), 3, 1,
    dimnames = list(c("1", "2", "3"), "ch")
  )
  expect_equal(
    capture_output(utf8_print(x, quote = FALSE, right = TRUE)),
    capture_output(print(x, quote = FALSE, right = TRUE))
  )
  expect_equal(
    capture_output(utf8_print(x, quote = TRUE, right = TRUE)),
    capture_output(print(x, quote = TRUE, right = TRUE))
  )
})


test_that("'utf8_print' does not need a gap at the end", {
  w <- 80
  x <- cbind(
    x = paste0(rep("x", 10), collapse = ""),
    y = paste0(rep("y", w - 13 - 5 - 2), collapse = "")
  )
  expect_equal(length(strsplit(
    capture_output(utf8_print(x)),
    "\n"
  )[[1]]), 2)
})


test_that("'utf8_print' wraps correctly", {
  w <- 80
  half <- floor(w / 2)
  d <- cbind(
    x = c("X", paste(rep("x", 2 * w), collapse = "")),
    y = c("Y", paste(rep("y", half + 1), collapse = "")),
    z = c("Z", paste(rep("z", half + 1), collapse = "")),
    a = 1:2,
    b = 3:4,
    c = 5:6
  )

  expect_equal(
    capture_output(utf8_print(d, chars = 1000, quote = FALSE)),
    capture_output(print(d, quote = FALSE))
  )

  d2 <- cbind(x = paste(rep("x", w - 2), collapse = ""), y = "y", z = "z")
  expect_equal(
    capture_output(utf8_print(d2, chars = 1000, quote = FALSE)),
    capture_output(print(d2, quote = FALSE))
  )

  expect_equal(
    capture_output(utf8_print(d2[, c(2, 1, 3), drop = FALSE],
      chars = 1000
    )),
    capture_output(print(d2[, c(2, 1, 3), drop = FALSE]))
  )

  expect_equal(
    capture_output(utf8_print(d2[, c(2, 3, 1), drop = FALSE],
      chars = 1000
    )),
    capture_output(print(d2[, c(2, 3, 1), drop = FALSE]))
  )

  d3 <- as.matrix(data.frame(
    x = "X", y = "Y", z = "Z",
    row.names = paste(rep("x", w), collapse = ""),
    stringsAsFactors = FALSE
  ))
  expect_equal(
    capture_output(utf8_print(d3, quote = FALSE)),
    capture_output(print(d3, quote = FALSE))
  )

  d4 <- as.matrix(data.frame(
    x = "X", y = "Y", z = "Z",
    row.names = paste(rep("x", w - 1), collapse = ""),
    stringsAsFactors = FALSE
  ))
  expect_equal(
    capture_output(utf8_print(d4, quote = FALSE)),
    capture_output(print(d4, quote = FALSE))
  )

  d5 <- as.matrix(data.frame(
    x = "X", y = "Y", z = "Z",
    row.names = paste(rep("x", w + 1), collapse = ""),
    stringsAsFactors = FALSE
  ))
  expect_equal(
    capture_output(utf8_print(d5, quote = FALSE)),
    capture_output(print(d5, quote = FALSE))
  )
})


chartype_matrix <- function() {
  chars <- character()
  desc <- character()

  chars[1] <- "\u0001\u001f"
  desc[1] <- "C0 control code"
  chars[2] <- "\a\b\f\n\r\t"
  desc[2] <- "Named control code"
  chars[3] <- "abcdefuvwxyz"
  desc[3] <- "ASCII"
  chars[4] <- "\u0080\u009f"
  desc[4] <- "C1 control code"

  chars[5] <- paste0(
    "\u00a0\u00a1\u00a2\u00a3\u00a4\u00a5",
    "\u00fa\u00fb\u00fc\u00fd\u00fe\u00ff"
  )
  desc[5] <- "Latin-1"

  chars[6] <- paste0(
    "\u0100\u0101\u0102\u0103\u0104\u0105",
    "\u0106\u0107\u0108\u0109\u010a\u010b"
  )
  desc[6] <- "Unicode"

  chars[7] <- "\uff01\uff02\uff03\uff04\uff05\uff06"
  desc[7] <- "Unicode wide"

  chars[8] <- "\ue00\u2029"
  desc[8] <- "Unicode control"

  chars[9] <- paste0(
    "x\u00adx\u200bx\u200cx\u200dx\u200ex\u200f",
    "x\u034fx\ufeffx", intToUtf8(0xE0001), "x",
    intToUtf8(0xE0020), "x", intToUtf8(0xE01EF), "x"
  )
  desc[9] <- "Unicode ignorable"

  chars[10] <- paste0(
    "a\u0300a\u0301a\u0302a\u0303a\u0304a\u0305",
    "a\u0306a\u0307a\u0308a\u0309a\u030aa\u030b"
  )
  desc[10] <- "Unicode mark"

  chars[11] <- paste0(
    intToUtf8(0x1F600), intToUtf8(0x1F601),
    intToUtf8(0x1F602), intToUtf8(0x1F603),
    intToUtf8(0x1F604), intToUtf8(0x1F483)
  )
  desc[11] <- "Emoji"

  chars[12] <- paste0("x", intToUtf8(0x10ffff), "x")
  desc[12] <- "Unassigned"

  chars[13] <- "\xfd\xfe\xff"
  desc[13] <- "Invalid"

  Encoding(chars) <- "UTF-8"

  x <- cbind(chars, desc)
  rownames(x) <- seq_len(nrow(x))
  x
}


test_that("'utf8_print' handles Unicode correctly", {
  # R can't print all UTF-8 on windows:
  # https://stat.ethz.ch/pipermail/r-devel/2017-June/074556.html
  skip_on_os("windows")
  local_ctype("UTF-8")

  x <- chartype_matrix()
  actual <- strsplit(
    capture_output(utf8_print(x, quote = FALSE)),
    "\n"
  )[[1]]
  Encoding(actual) <- "UTF-8"

  expected <- c(
    "   chars        desc              ",
    "1  \\u0001\\u001f C0 control code   ",
    "2  \\a\\b\\f\\n\\r\\t Named control code",
    "3  abcdefuvwxyz ASCII             ",
    "4  \\u0080\\u009f C1 control code   ",
    paste0("5  ", x[5, "chars"], " Latin-1           "),
    paste0("6  ", x[6, "chars"], " Unicode           "),
    "7  \uff01\uff02\uff03\uff04\uff05\uff06 Unicode wide      ",
    "8  \\u0e00\\u2029 Unicode control   ",
    "9  xxxxxxxxxxxx Unicode ignorable ",
    paste0("10 ", x[10, "chars"], " Unicode mark      "),
    paste0("11 ", paste(intToUtf8(0x1F600), intToUtf8(0x1F601),
      intToUtf8(0x1F602), intToUtf8(0x1F603),
      intToUtf8(0x1F604), intToUtf8(0x1F483), "",
      sep = "\u200b"
    ), " Emoji             "),
    "12 x\\U0010ffffx Unassigned        ",
    "13 \\xfd\\xfe\\xff Invalid           "
  )
  Encoding(expected) <- "UTF-8"

  expect_equal(actual, expected)
})


test_that("'utf8_print' works in C locale", {
  x <- chartype_matrix()

  # https://github.com/r-lib/testthat/issues/1285
  with_ctype("C", {
    actual <- strsplit(
      capture_output(utf8_print(x,
        chars = 1000,
        quote = FALSE
      )),
      "\n"
    )[[1]]
  })

  expected <- c(
    "   chars                                                                                     ",
    "1  \\u0001\\u001f                                                                              ",
    "2  \\a\\b\\f\\n\\r\\t                                                                              ",
    "3  abcdefuvwxyz                                                                              ",
    "4  \\u0080\\u009f                                                                              ",
    "5  \\u00a0\\u00a1\\u00a2\\u00a3\\u00a4\\u00a5\\u00fa\\u00fb\\u00fc\\u00fd\\u00fe\\u00ff                  ",
    "6  \\u0100\\u0101\\u0102\\u0103\\u0104\\u0105\\u0106\\u0107\\u0108\\u0109\\u010a\\u010b                  ",
    "7  \\uff01\\uff02\\uff03\\uff04\\uff05\\uff06                                                      ",
    "8  \\u0e00\\u2029                                                                              ",
    "9  x\\u00adx\\u200bx\\u200cx\\u200dx\\u200ex\\u200fx\\u034fx\\ufeffx\\U000e0001x\\U000e0020x\\U000e01efx",
    "10 a\\u0300a\\u0301a\\u0302a\\u0303a\\u0304a\\u0305a\\u0306a\\u0307a\\u0308a\\u0309a\\u030aa\\u030b      ",
    "11 \\U0001f600\\U0001f601\\U0001f602\\U0001f603\\U0001f604\\U0001f483                              ",
    "12 x\\U0010ffffx                                                                              ",
    "13 \\xfd\\xfe\\xff                                                                              ",
    "   desc              ",
    "1  C0 control code   ",
    "2  Named control code",
    "3  ASCII             ",
    "4  C1 control code   ",
    "5  Latin-1           ",
    "6  Unicode           ",
    "7  Unicode wide      ",
    "8  Unicode control   ",
    "9  Unicode ignorable ",
    "10 Unicode mark      ",
    "11 Emoji             ",
    "12 Unassigned        ",
    "13 Invalid           "
  )

  expect_equal(actual, expected)
})

if (utils::packageVersion("testthat") <= "1.0.2") {
  capture_output <- function(code, print = FALSE, width = 80) {
    oldwidth <- getOption("width")
    if (width != oldwidth) {
      skip_if_not(with(R.Version(), paste(major, minor, sep = "."))
      >= "3.4.0", "Setting width fails on R < 3.4.0")

      options(width = width)
      on.exit(options(width = oldwidth), add = TRUE)
    }
    testthat::capture_output(code, print)
  }
}

context("utf8_print")

test_that("'utf8_print' can print unicode", {
    skip_on_os("windows")
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    x <-  c("\u0100\u0101\u0102\u0103\u0104\u0105",
            "\u0106\u0107\u0108\u0109\u010a\u010b")

    expect_equal(capture_output(utf8_print(x)),
                 paste("[1] \"\u0100\u0101\u0102\u0103\u0104\u0105\"",
                       "\"\u0106\u0107\u0108\u0109\u010a\u010b\""))
})

test_that("'utf8_print' works with unnamed character vectors", {
    x <- as.character(1:100)

    expect_equal(capture_output(utf8_print(x)),
                 capture_output(print(x)))

    expect_equal(capture_output(utf8_print(x[1:96])),
                 capture_output(print(x[1:96])))

    expect_equal(capture_output(utf8_print(x[1:7])),
                 capture_output(print(x[1:7])))
})


test_that("'utf8_print' works with named character vectors", {
    x <- as.character(10 + 1:26)
    names(x) <- letters

    # left align names
    xr <- x
    names(xr) <- format(names(x), aligh="left", width = 4)
    actual <- strsplit(capture_output(utf8_print(x)), "\n")[[1]]
    expected <- strsplit(capture_output(print(xr)), "\n")[[1]]
    expect_equal(paste(actual, ""), expected)

    actual <- strsplit(capture_output(utf8_print(x[1:16])), "\n")[[1]]
    expected <- strsplit(capture_output(print(xr[1:16])), "\n")[[1]]
    expect_equal(paste(actual, ""), expected)

    actual <- strsplit(capture_output(utf8_print(x[1:4])), "\n")[[1]]
    expected <- strsplit(capture_output(print(xr[1:4])), "\n")[[1]]
    expect_equal(paste(actual, ""), expected)
})


test_that("'utf8_print' can use the 'max' argument for unnamed vectors", {
    x <- as.character(1:100)

    expect_equal(capture_output(utf8_print(x, max = 0), width = 80),
                 " [ reached getOption(\"max.print\") -- omitted 100 entries ]")

    expect_equal(capture_output(utf8_print(x, max = 100), width = 80),
                 capture_output(utf8_print(x), width = 80))

    lines <- strsplit(capture_output(utf8_print(x, max = 20), width = 80),
                      "\n")[[1]]
    expect_equal(length(lines), 3)
    expect_equal(lines[[3]],
                 " [ reached getOption(\"max.print\") -- omitted 80 entries ]")
})


test_that("'utf8_print' can use the 'max' argument for named vectors", {
    x <- as.character(1:260)
    names(x) <- rep(letters, 10)

    expect_equal(capture_output(utf8_print(x, max = 0), width = 80),
                 " [ reached getOption(\"max.print\") -- omitted 260 entries ]")

    expect_equal(capture_output(utf8_print(x, max = 260), width = 80),
                 capture_output(utf8_print(x), width = 80))

    lines <- strsplit(capture_output(utf8_print(x, max = 20), width = 80),
                      "\n")[[1]]
    expect_equal(length(lines), 5)
    expect_equal(lines[[5]],
                 " [ reached getOption(\"max.print\") -- omitted 240 entries ]")
})


test_that("'utf8_print' can print empty vectors", {
    expect_equal(capture_output(utf8_print(character())), "character(0)")
    expect_equal(capture_output(utf8_print(array(character(), 0))), "character(0)")
})


test_that("'utf8_print' can print matrices", {
    x1 <- matrix(letters, 13, 2)

    x2 <- matrix(letters, 13, 2)
    rownames(x2) <- LETTERS[1:13]

    x3 <- matrix(letters, 13, 2)
    colnames(x3) <- c("x", "y")

    x4 <- matrix(letters, 13, 2)
    rownames(x4) <- LETTERS[1:13]
    colnames(x4) <- c("x", "y")

    expect_equal(capture_output(utf8_print(x1)),
                 capture_output(print(x1)))

    expect_equal(capture_output(utf8_print(x2)),
                 capture_output(print(x2)))

    expect_equal(capture_output(utf8_print(x3)),
                 capture_output(print(x3)))

    expect_equal(capture_output(utf8_print(x4)),
                 capture_output(print(x4)))
})


test_that("'utf8_print' can print empty matrices", {
    x1 <- matrix(character(), 10, 0)
    x2 <- matrix(character(), 0, 10)
    x3 <- matrix(character(), 0, 0)

    expect_equal(paste0("     \n", capture_output(utf8_print(x1))),
                 capture_output(print(x1)))

    expect_equal(paste0("     ", capture_output(utf8_print(x2))),
                 capture_output(print(x2)))

    expect_equal(capture_output(utf8_print(x3)),
                 capture_output(print(x3)))
})


test_that("'utf8_print' can print arrays", {
    x <- array(as.character(1:24), c(2,3,4,5))

    expect_equal(capture_output(utf8_print(x)),
                 capture_output(print(x)))

    x2 <- x
    dimnames(x2) <- list(letters[1:2], letters[3:5], letters[6:9],
                         letters[10:14])

    expect_equal(capture_output(utf8_print(x2)),
                 capture_output(print(x2)))
})


test_that("'utf8_print' can print empty arrays", {
    expect_equal(capture_output(utf8_print(array(character(), c(2,3,0)))),
                 "<2 x 3 x 0 array>")

    expect_equal(capture_output(utf8_print(array(character(), c(2,0,3)))),
                 "<2 x 0 x 3 array>")

    expect_equal(capture_output(utf8_print(array(character(), c(0,2,3)))),
                 "<0 x 2 x 3 array>")
})


test_that("'utf8_print' can print quotes", {
    expect_equal(capture_output(utf8_print('"')),
                 capture_output(print('"')))

    expect_equal(capture_output(utf8_print('"', quote = FALSE)),
                 capture_output(print('"', quote = FALSE)))
})


test_that("'utf8_print' can handle NA", {
    expect_equal(capture_output(utf8_print(NA_character_)),
                 capture_output(print(NA_character_)))
    expect_equal(capture_output(utf8_print(NA_character_, quote = FALSE)),
                 capture_output(print(NA_character_, quote = FALSE)))
})


test_that("'utf8_print' can handle NA names", {
    x <- matrix("hello", 1, 1, dimnames=list(NA,NA))
    expect_equal(capture_output(utf8_print(x)),
                 capture_output(print(x)))
    expect_equal(capture_output(utf8_print(x, na.print = "foo")),
                 capture_output(print(x, na.print = "foo")))
})


test_that("'utf8_print' can right justify", {
    x <- matrix(c("a", "ab", "abc"), 3, 1,
                dimnames = list(c("1", "2", "3"), "ch"))
    expect_equal(capture_output(utf8_print(x, quote = FALSE, right = TRUE)),
                 capture_output(print(x, quote = FALSE, right = TRUE)))
    expect_equal(capture_output(utf8_print(x, quote = TRUE, right = TRUE)),
                 capture_output(print(x, quote = TRUE, right = TRUE)))
})


test_that("'utf8_print' does not need a gap at the end", {
    w <- 80
    x <- cbind(x = paste0(rep("x", 10), collapse=""),
               y = paste0(rep("y", w - 13 - 5 - 2), collapse=""))
    expect_equal(length(strsplit(capture_output(utf8_print(x)),
                                 "\n")[[1]]), 2)
})


test_that("'utf8_print' wraps correctly", {
    w <- 80
    half <- floor(w / 2)
    d <- cbind(x = c("X", paste(rep("x", 2 * w), collapse="")),
               y = c("Y", paste(rep("y", half + 1), collapse="")),
               z = c("Z", paste(rep("z", half + 1), collapse="")),
               a = 1:2,
               b = 3:4,
               c = 5:6)

    expect_equal(capture_output(utf8_print(d, chars = 1000, quote = FALSE)),
                 capture_output(print(d, quote = FALSE)))

    d2 <- cbind(x = paste(rep("x", w - 2), collapse=""), y = "y", z = "z")
    expect_equal(capture_output(utf8_print(d2, chars = 1000, quote = FALSE)),
                 capture_output(print(d2, quote = FALSE)))

    expect_equal(capture_output(utf8_print(d2[,c(2,1,3), drop = FALSE],
                                           chars = 1000)),
                 capture_output(print(d2[,c(2,1,3), drop = FALSE])))

    expect_equal(capture_output(utf8_print(d2[,c(2,3,1), drop = FALSE],
                                           chars = 1000)),
                 capture_output(print(d2[,c(2,3,1), drop = FALSE])))

    d3 <- as.matrix(data.frame(x = "X", y = "Y", z = "Z",
                               row.names = paste(rep("x", w), collapse=""),
                               stringsAsFactors = FALSE))
    expect_equal(capture_output(utf8_print(d3, quote = FALSE)),
                 capture_output(print(d3, quote = FALSE)))

    d4 <- as.matrix(data.frame(x = "X", y = "Y", z = "Z",
                               row.names = paste(rep("x", w - 1), collapse=""),
                               stringsAsFactors = FALSE))
    expect_equal(capture_output(utf8_print(d4, quote = FALSE)),
                 capture_output(print(d4, quote = FALSE)))

    d5 <- as.matrix(data.frame(x = "X", y = "Y", z = "Z",
                               row.names = paste(rep("x", w + 1), collapse=""),
                               stringsAsFactors = FALSE))
    expect_equal(capture_output(utf8_print(d5, quote = FALSE)),
                 capture_output(print(d5, quote = FALSE)))
})


chartype_matrix <- function()
{
    chars <- character()
    desc <- character()

    chars[1] <- "\u0001\u001f"; desc[1] <- "C0 control code"
    chars[2] <- "\a\b\f\n\r\t"; desc[2] <- "Named control code"
    chars[3] <- "abcdefuvwxyz"; desc[3] <- "ASCII"
    chars[4] <- "\u0080\u009f"; desc[4] <- "C1 control code"

    chars[5] <- paste0("\u00a0\u00a1\u00a2\u00a3\u00a4\u00a5",
                       "\u00fa\u00fb\u00fc\u00fd\u00fe\u00ff")
    desc[5] <- "Latin-1"

    chars[6] <- paste0("\u0100\u0101\u0102\u0103\u0104\u0105",
                       "\u0106\u0107\u0108\u0109\u010a\u010b")
    desc[6] <- "Unicode"

    chars[7] <- "\uff01\uff02\uff03\uff04\uff05\uff06"
    desc[7] <- "Unicode wide"

    chars[8] <- "\ue00\u2029"
    desc[8] <- "Unicode control"

    chars[9] <- paste0("x\u00adx\u200bx\u200cx\u200dx\u200ex\u200f",
                       "x\u034fx\ufeffx", intToUtf8(0xE0001), "x",
                       intToUtf8(0xE0020), "x", intToUtf8(0xE01EF), "x")
    desc[9] <- "Unicode ignorable"

    chars[10] <- paste0("a\u0300a\u0301a\u0302a\u0303a\u0304a\u0305",
                        "a\u0306a\u0307a\u0308a\u0309a\u030aa\u030b")
    desc[10] <- "Unicode mark"

    chars[11] <- paste0(intToUtf8(0x1F600), intToUtf8(0x1F601),
                        intToUtf8(0x1F602), intToUtf8(0x1F603),
                        intToUtf8(0x1F604), intToUtf8(0x1F483))
    desc[11] <- "Emoji"

    chars[12] <- paste0("x", intToUtf8(0x10ffff), "x")
    desc[12] <- "Unassigned"

    chars[13] <- "\xfd\xfe\xff"
    desc[13] <- "Invalid"

    Encoding(chars) <- "UTF-8"

    x <- cbind(chars, desc)
    rownames(x) <- seq_len(nrow(x))
    x
}


test_that("'utf8_print' handles Unicode correctly", {
    # R can't print all UTF-8 on windows:
    # https://stat.ethz.ch/pipermail/r-devel/2017-June/074556.html
    skip_on_os("windows")
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    x <- chartype_matrix()
    actual <- strsplit(capture_output(utf8_print(x, quote = FALSE)),
                       "\n")[[1]]
    Encoding(actual) <- "UTF-8"

    expected <- c(
        "   chars        desc              ",
        "1  \\u0001\\u001f C0 control code   ",
        "2  \\a\\b\\f\\n\\r\\t Named control code",
        "3  abcdefuvwxyz ASCII             ",
        "4  \\u0080\\u009f C1 control code   ",
        paste0("5  ", x[5, "chars"], " Latin-1           "),
        paste0("6  ", x[6, "chars"], " Unicode           "),
        "7  \uff01\uff02\uff03\uff04\uff05\uff06 Unicode wide      ",
        "8  \\u0e00\\u2029 Unicode control   ",
        "9  xxxxxxxxxxxx Unicode ignorable ",
        paste0("10 ", x[10, "chars"], " Unicode mark      "),
        paste0("11 ", paste(intToUtf8(0x1F600), intToUtf8(0x1F601),
                            intToUtf8(0x1F602), intToUtf8(0x1F603),
                            intToUtf8(0x1F604), intToUtf8(0x1F483), "",
                            sep = "\u200b"), " Emoji             "),
        "12 x\\U0010ffffx Unassigned        ",
        "13 \\xfd\\xfe\\xff Invalid           ")
    Encoding(expected) <- "UTF-8"

    expect_equal(actual, expected)
})


test_that("'utf8_print' works in C locale", {
    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    x <- chartype_matrix()
    actual <- strsplit(capture_output(utf8_print(x, chars = 1000,
                                                 quote = FALSE)),
                       "\n")[[1]]

    expected <- c(
    "   chars                                                                                     ",
    "1  \\u0001\\u001f                                                                              ",
    "2  \\a\\b\\f\\n\\r\\t                                                                              ",
    "3  abcdefuvwxyz                                                                              ",
    "4  \\u0080\\u009f                                                                              ",
    "5  \\u00a0\\u00a1\\u00a2\\u00a3\\u00a4\\u00a5\\u00fa\\u00fb\\u00fc\\u00fd\\u00fe\\u00ff                  ",
    "6  \\u0100\\u0101\\u0102\\u0103\\u0104\\u0105\\u0106\\u0107\\u0108\\u0109\\u010a\\u010b                  ",
    "7  \\uff01\\uff02\\uff03\\uff04\\uff05\\uff06                                                      ",
    "8  \\u0e00\\u2029                                                                              ",
    "9  x\\u00adx\\u200bx\\u200cx\\u200dx\\u200ex\\u200fx\\u034fx\\ufeffx\\U000e0001x\\U000e0020x\\U000e01efx",
    "10 a\\u0300a\\u0301a\\u0302a\\u0303a\\u0304a\\u0305a\\u0306a\\u0307a\\u0308a\\u0309a\\u030aa\\u030b      ",
    "11 \\U0001f600\\U0001f601\\U0001f602\\U0001f603\\U0001f604\\U0001f483                              ",
    "12 x\\U0010ffffx                                                                              ",
    "13 \\xfd\\xfe\\xff                                                                              ",
    "   desc              ",
    "1  C0 control code   ",
    "2  Named control code",
    "3  ASCII             ",
    "4  C1 control code   ",
    "5  Latin-1           ",
    "6  Unicode           ",
    "7  Unicode wide      ",
    "8  Unicode control   ",
    "9  Unicode ignorable ",
    "10 Unicode mark      ",
    "11 Emoji             ",
    "12 Unassigned        ",
    "13 Invalid           ")

    expect_equal(actual, expected)
})

switch_ctype <- function(mode = c("C", "UTF-8"))
{
    mode <- match.arg(mode)

    if (mode == "UTF-8") {
        sysname <- Sys.info()[["sysname"]]
        if (sysname == "Windows") {
            ctype <- "English_United States.1252"
        } else if (sysname == "Darwin") {
            ctype <- "UTF-8"
        } else {
            ctype <- "en_US.utf8"
        }
    } else {
        ctype <- "C"
    }

    ctype0 <- Sys.getlocale("LC_CTYPE")
    suppressWarnings({
        Sys.setlocale("LC_CTYPE", ctype)
    })
    if (Sys.getlocale("LC_CTYPE") != ctype) {
        skip(paste0("Cannot change locale to '", ctype, "'"))
    }
    if (mode == "UTF-8" && !output_utf8()) {
        skip("Cannot change to UTF-8 output")
    }

    ctype0
}

context("utf8_width")

test_that("'utf8_width' computes widths correctly", {
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_width(c("hello", "\u200b", "\u22ee", "\u6027"),
                            encode = FALSE),
                 c(5, 0, 1, 2))
})


test_that("'utf8_width' computes widths for extended unicode correctly", {
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))
    skip_on_os("windows") # no extended Unicode

    expect_equal(utf8_width(intToUtf8(0x1f642), encode = FALSE), 2)
})




test_that("'utf8_width' gives NA for non-ASCII in C locale", {
    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    x <- c("hello", "\u200b", "\u22ee", "\u6027", intToUtf8(0x1f642))
    expect_equal(utf8_width(x, encode = FALSE), c(5, NA, NA, NA, NA))
})


test_that("'utf8_width' keeps names", {
    expect_equal(utf8_width(c(a = "hello", b = "you")),
                 c(a = 5, b = 3))
})


test_that("'utf8_width' gives NA for control characters", {
    expect_equal(utf8_width(c("\u0001", "\a", "new\nline"),
                            encode = FALSE),
                            c(NA_integer_, NA_integer_, NA_integer_))
})


test_that("'utf8_width' gives NA for invalid data", {
    x <- c("a", "b", "\xff", "abc\xfe")
    Encoding(x) <- "UTF-8"
    expect_equal(utf8_width(x, encode = FALSE), c(1, 1, NA, NA))
})


test_that("'utf8_width' gives width 1 for quotes", {
    expect_equal(utf8_width("\"", encode = TRUE), 1)
    expect_equal(utf8_width("\"", encode = FALSE), 1)
})


test_that("'utf8_width' gives correct with in C locale", {
    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))
    x <- intToUtf8(c(0x1F487, 0x200D, 0x2642, 0xFE0F))
    expect_equal(utf8_width(x), nchar(utf8_encode(x)))
})

context("utf8_encode")


test_that("'utf8_encode' can encode an ASCII string", {
    expect_equal(utf8_encode("hello"), "hello")
})


test_that("'utf8_encode' can encode NULL or an NA string", {
    expect_equal(utf8_encode(NULL), NULL)
    expect_equal(utf8_encode(NA_character_), NA_character_)
})


test_that("'utf8_encode' preserves attributes", {
    x <- matrix(LETTERS, 2, 13)
    x[1,4] <- "\xa4"
    Encoding(x) <- "latin1"
    dimnames(x) <- list(c("foo", "bar"), as.character(1:13))
    class(x) <- "my_class"

    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_encode(x), enc2utf8(x))
})


test_that("'utf8_encode' can encode basic Unicode", {
    x <- "\u200b"
    Encoding(x) <- "UTF-8"

    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_encode(x), x)
})


test_that("'utf8_encode' can encode extended Unicode", {
    skip_on_os("windows") # no extended Unicode

    x <- intToUtf8(0x0001f60d)
    Encoding(x) <- "UTF-8"

    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_encode(x), x)
})


test_that("'utf8_encode' can handle ASCII escapes", {
    x <- "\x01\a\b\f\n\r\t\v\x7f"
    expect_equal(utf8_encode(x), "\\u0001\\a\\b\\f\\n\\r\\t\\v\\u007f")
})

test_that("'utf8_encode' can handle invalid UTF-8", {
    x <- "\xfe"
    Encoding(x) <- "bytes"
    expect_equal(utf8_encode(x), "\\xfe")
})

test_that("'utf8_encode' can handle bytes", {
    x <- "\x01\a\b\f\n\r\t\v\x7f\x80\xff"
    Encoding(x) <- "bytes"
    expect_equal(utf8_encode(x),
                 "\\x01\\a\\b\\f\\n\\r\\t\\v\\x7f\\x80\\xff")
})

test_that("'utf8_encode' can handle latin-1", {
    x <- "her \xa320"
    Encoding(x) <- "latin1"

    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_encode(x), "her \\u00a320")

    switch_ctype("UTF-8")
    expect_equal(utf8_encode(x), "her \u00a320")
})


test_that("'utf8_encode' can handle bytes", {
    x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
    Encoding(x) <- c("UTF-8", "UTF-8", "bytes")

    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    y <- c("fa\u00e7ile", "fa\\xe7ile", "fa\\xc3\\xa7ile")
    Encoding(y) <- c("UTF-8", "UTF-8", "bytes")

    expect_equal(utf8_encode(x), y)
})


test_that("'utf8_encode escapes controls in UTF-8 text", {
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    x <- '\n\u2026'; Encoding(x) <- "UTF-8"
    y <- '\\n\u2026'; Encoding(y) <- "UTF-8"
    expect_equal(utf8_encode(x), y)
})


test_that("'utf8_encode' can quote", {
    x <- c("abcde", "x", "123", "\"", "'")
    expect_equal(utf8_encode(x, quote = FALSE), encodeString(x, quote = ""))
    expect_equal(utf8_encode(x, quote = TRUE), encodeString(x, quote = '"'))
})


test_that("'utf8_encode' ignores justify if width = 0", {
    x <- c("abcde", "x", "123", "\"", "'")
    expect_equal(utf8_encode(x, width = 0, justify = "none"),
                 encodeString(x, width = 0, justify = "none"))
    expect_equal(utf8_encode(x, width = 0, justify = "left"),
                 encodeString(x, width = 0, justify = "left"))
    expect_equal(utf8_encode(x, width = 0, justify = "centre"),
                 encodeString(x, width = 0, justify = "centre"))
    expect_equal(utf8_encode(x, width = 0, justify = "right"),
                 encodeString(x, width = 0, justify = "right"))
})


test_that("'utf8_encode' can justify", {
    x <- c("abcde", "x", "123", "\"", "'")
    expect_equal(utf8_encode(x, width = NULL, justify = "none"),
                 encodeString(x, width = NULL, justify = "none"))
    expect_equal(utf8_encode(x, width = NULL, justify = "left"),
                 encodeString(x, width = NULL, justify = "left"))
    expect_equal(utf8_encode(x, width = NULL, justify = "centre"),
                 encodeString(x, width = NULL, justify = "centre"))
    expect_equal(utf8_encode(x, width = NULL, justify = "right"),
                 encodeString(x, width = NULL, justify = "right"))
})


test_that("'utf8_encode' can justify and quote", {
    expect_equal(utf8_encode(c("1", "10", "100"), width = NULL, quote = TRUE),
                 encodeString(c("1", "10", "100"), width = NULL, quote = '"'))
})

context("utf8_format")


test_that("'format' can handle short text", {
    raw <- c(NA, "", "a", "foo", "short text")

    expect_equal(utf8_format(raw, justify = "none", na.print = "NA"),
                 format(raw, justify = "none"))

    expect_equal(utf8_format(raw, justify = "left", na.print = "NA"),
                 format(raw, justify = "left"))

    expect_equal(utf8_format(raw, justify = "centre", na.print = "NA"),
                 format(raw, justify = "centre"))

    expect_equal(utf8_format(raw, justify = "right", na.print = "NA"),
                 format(raw, justify = "right"))
})


test_that("'format' can handle long text in Unicode locale", {
    raw    <- c(NA, "", "a", "ab", "foo", "food",     "short text",
                "\u6027", "\u6027\u6027", "\u6027?")
    Encoding(raw) <- "UTF-8"

    short  <- c(NA, "", "a", "ab", "fo\u2026", "fo\u2026", "sh\u2026",
                "\u6027", "\u6027\u2026", "\u6027\u2026")
    Encoding(short) <- "UTF-8"

    rshort <- c(NA, "", "a", "ab", "\u2026oo", "\u2026od", "\u2026xt",
                "\u6027", "\u2026\u6027", "\u2026?")
    Encoding(rshort) <- "UTF-8"

    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))
    skip_on_os("windows") # windows can't format \u6027

    expect_equal(utf8_format(raw, chars = 2, justify = "none", na.print = "NA"),
                 format(short, justify = "none"))

    expect_equal(utf8_format(raw, chars = 2, justify = "left", na.print = "NA"),
                 format(short, justify = "left"))

    expect_equal(utf8_format(raw, chars = 2, justify = "centre", na.print = "NA"),
                 format(short, justify = "centre"))

    expect_equal(utf8_format(raw, chars = 2, justify = "right", na.print = "NA"),
                 format(rshort, justify = "right"))
})


test_that("'format' can handle long text in UTF-8 locale, part 2", {
    raw    <- c(NA, "", "a", "\n", "ab", "foo", "food",     "short text",
                "\u6027", "\u6027\u6027", "\u6027?")
    short  <- c(NA, "", "a", "\u2026", "a\u2026", "f\u2026", "f\u2026",
                "s\u2026", "\u2026", "\u2026", "\u2026")
    rshort <- c(NA, "", "a", "\u2026", "\u2026b", "\u2026o", "\u2026d",
                "\u2026t", "\u2026", "\u2026", "\u2026?")

    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_format(raw, chars = 1, justify = "none",
                             na.encode = FALSE),
                 format(short, justify = "none", na.encode = FALSE))

    expect_equal(utf8_format(raw, chars = 1, justify = "left", na.encode = FALSE),
                 format(short, justify = "left", na.encode = FALSE))

    expect_equal(utf8_format(raw, chars = 1, justify = "centre", na.encode = FALSE),
                 format(short, justify = "centre", na.encode = FALSE))

    expect_equal(utf8_format(raw, chars = 1, justify = "right",
                             na.encode = FALSE),
                 format(rshort, justify = "right", na.encode = FALSE))
})


test_that("'format' can handle long text in C locale", {
    #            6         7          8           9            10
    raw    <- c("\u6027", "?\u6027", "\n\u6027", "?\n\u6027", "\u0001\u6027",
                          "\u6027?", "\u6027\n", "\u6027?\n", "\u6027\u0001")
    short  <- c("\\u6027", "?\\u6027", "\\n\\u6027", "?\\n...",     "\\u0001...",
                          "\\u6027?", "\\u6027\\n", "\\u6027?...", "\\u6027...")
    rshort <- c("\\u6027", "?\\u6027", "\\n\\u6027", "...\\n\\u6027",
                 "...\\u6027", "\\u6027?", "\\u6027\\n", "...?\\n",
                  "...\\u0001")

    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_encode(utf8_format(raw, chars = 8, justify = "none")),
                 format(short, justify = "none"))

    left <- utf8_encode(utf8_format(raw, chars = 8, justify = "left"))
    expect_equal(sub("\\s+$", "", left), short)
    expect_equal(as.numeric(nchar(left)), rep(10, length(raw)))

    centre <- utf8_encode(utf8_format(raw, chars = 8, justify = "centre"))
    expect_equal(sub("^\\s+", "", sub("\\s+$", "", centre)), short)
    expect_equal(as.numeric(nchar(centre)), rep(10, length(raw)))

    right <- utf8_encode(utf8_format(raw, chars = 8, justify = "right"))
    expect_equal(sub("^\\s+", "", right), rshort)
    expect_equal(as.numeric(nchar(right)), rep(11, length(raw)))
})


test_that("'format' can handle high code points in C locale", {
    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    raw <- c(intToUtf8(0x00010000), intToUtf8(0x0010ffff))

    expect_equal(utf8_format(raw, justify = "left"), raw)
    expect_equal(utf8_format(raw, justify = "right"), raw)
})


test_that("'format' can handle high code points in Unicode locale", {
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))
    skip_on_os("windows") # no Unicode above 0xFFFF on Windows

    raw   <- c(intToUtf8(0x00010000), intToUtf8(0x010ffff))
    left  <- c(paste0(intToUtf8(0x00010000), "         "), intToUtf8(0x010ffff))
    right <- c(paste0("         ", intToUtf8(0x00010000)), intToUtf8(0x010ffff))

    expect_equal(utf8_format(raw, justify = "left"), left)
    expect_equal(utf8_format(raw, justify = "right"), right)
})


test_that("'format' can handle ignorable code points", {
    raw <- "\u200B"

    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_format(raw, justify = "left"), raw)
    expect_equal(utf8_format(raw, justify = "centre"), raw)
    expect_equal(utf8_format(raw, justify = "right"), raw)

    switch_ctype("UTF-8")

    expect_equal(utf8_format(raw, justify = "left"), raw)
    expect_equal(utf8_format(raw, justify = "centre"), raw)
    expect_equal(utf8_format(raw, justify = "right"), raw)
})


test_that("'format' can handle marks", {
    raw <- "\u1e0d\u0307"

    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_format(raw, chars = 6, justify = "left"), "...")

    expect_equal(utf8_format(raw, chars = 6, justify = "centre"), "...")

    expect_equal(utf8_format(raw, chars = 5, justify = "right"), "...")

    switch_ctype("UTF-8")

    expect_equal(utf8_format(raw, chars = 1, justify = "left"), raw)
    expect_equal(utf8_format(raw, chars = 1, justify = "centre"), raw)
    expect_equal(utf8_format(raw, chars = 1, justify = "right"), raw)
})


test_that("'format' can handle UTF-8 'Other' codes", {
    raw <- "\u2072" # unassigned

    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    expect_equal(utf8_format(raw, justify = "left"), raw)
    expect_equal(utf8_format(raw, justify = "centre"), raw)
    expect_equal(utf8_format(raw, justify = "right"), raw)

    switch_ctype("UTF-8")

    expect_equal(utf8_format(raw, justify = "left"), raw)
    expect_equal(utf8_format(raw, justify = "centre"), raw)
    expect_equal(utf8_format(raw, justify = "right"), raw)
})


test_that("'format' can handle zero, or NULL chars", {
    raw <- "foo"

    ctype <- switch_ctype("C")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))
    
    expect_equal(as.character(utf8_format(raw, chars = 0, justify = "left")),
                 "...")
    expect_equal(as.character(utf8_format(raw, chars = 0, justify = "centre")),
                 "...")
    expect_equal(as.character(utf8_format(raw, chars = 0, justify = "right")),
                 "...")

    expect_equal(as.character(utf8_format(raw, chars = NULL, justify = "left")),
                 "foo")
    expect_equal(as.character(utf8_format(raw, chars = NULL,
                                          justify = "centre")),
                 "foo")
    expect_equal(as.character(utf8_format(raw, chars = NULL,
                                          justify = "right")),
                 "foo")
})


test_that("'format' can skip NA", {
    expect_equal(as.character(utf8_format(NA_character_, na.encode = FALSE)),
                 NA_character_)
})


test_that("'format' can set minimum width", {
    raw <- c("a", "ab", "abc")

    expect_equal(utf8_format(raw, justify = "none", width = 5),
                 format(raw, justify = "none", width = 5))
    expect_equal(utf8_format(raw, justify = "left", width = 5),
                 format(raw, justify = "left", width = 5))
    expect_equal(utf8_format(raw, justify = "centre", width = 5),
                 format(raw, justify = "centre", width = 5))
    expect_equal(utf8_format(raw, justify = "right", width = 5),
                 format(raw, justify = "right", width = 5))
})


test_that("'format' error for invalid justify", {
    expect_error(utf8_format("", justify = "wild"),
                 paste("'justify' must be one of the following:",
                       paste(dQuote(c("left", "right", "centre", "none")),
                             collapse = ", ")),
                 fixed = TRUE)
})


test_that("'format' error for invalid logicals", {
    expect_error(utf8_format("", trim = NA), "'trim' must be TRUE or FALSE",
                 fixed = TRUE)
    expect_error(utf8_format("", na.encode = NA),
                 "'na.encode' must be TRUE or FALSE", fixed = TRUE)
})


test_that("'format' error for invalid integers", {
    expect_error(utf8_format("", chars = "3"),
                 "'chars' must be integer-valued",
                 fixed = TRUE)
    expect_error(utf8_format("", width = "3"),
                 "'width' must be integer-valued",
                 fixed = TRUE)
})


test_that("'utf8_format' can handle invalid UTF-8", {
    input <- "\xff\xfe"
    Encoding(input) <- "UTF-8"
    x <- c("a", input)
    output <- input
    Encoding(output) <- "bytes"

    expect_equal(utf8_format(x, justify = "none"), c("a", output))
    expect_equal(utf8_format(x, justify = "left"), c("a       ", output))
    expect_equal(utf8_format(x, justify = "centre"), c("   a    ", output))
    expect_equal(utf8_format(x, justify = "right"), c("       a", output))
})


test_that("'utf8_format' can handle latin1 text", {
    x <- "fa\xE7ile"
    Encoding(x) <- "latin1"
    y <- iconv(x, "latin1", "UTF-8")
    expect_equal(utf8_format(x), y)
})


test_that("'utf8_format' can quote", {
    expect_equal(utf8_format(c("a", "abc"), quote = TRUE),
                 c('a  ', 'abc'))

    expect_equal(utf8_format(c("a", "abc"), quote = TRUE, justify = "centre"),
                 c(' a ', 'abc'))

    expect_equal(utf8_format(c("a", "abc"), quote = TRUE, justify = "right"),
                 c('  a', 'abc'))
})


test_that("'utf8_format' can handle quotes", {
    expect_equal(utf8_format('"'), '"')
    expect_equal(utf8_format('"', quote = TRUE), '"')
    expect_equal(utf8_format('"', justify = "right"), '"')
    expect_equal(utf8_format('"', justify = "right", quote = TRUE), '"')
})


test_that("'utf8_format' works on bytes", {
    x <- "fa\xC3\xA7ile"
    Encoding(x) <- "bytes"

    l <- utf8_format(x, justify = "left")
    expect_equal(Encoding(l), "bytes")

    r <- utf8_format(x, justify = "right")
    expect_equal(Encoding(r), "bytes")

    Encoding(x) <- Encoding(l) <- Encoding(r) <- "UTF-8"
    expect_equal(l, x)
    expect_equal(r, x)
})


test_that("'utf8_format' can right justify", {
    x <- " (3322 rows total)"
    expect_equal(utf8_format(x, width = 79, justify = "right"),
                 format(x, width = 79, justify = "right"))
})


test_that("'utf8_format' use ... ellipsis for bytes", {
    ctype <- switch_ctype("UTF-8")
    on.exit(Sys.setlocale("LC_CTYPE", ctype))

    x <- "fa\xC3\xA7ile"
    Encoding(x) <- "bytes"
    y <- "fa\xC3..."
    Encoding(y) <- "bytes"

    expect_equal(utf8_format(x, chars = 6), y)
})

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


with_rethrow <- function(expr) {
  parentcall <- sys.call(-1)
  eval(
    envir = parent.frame(),
    withCallingHandlers(expr,
      error = function(e, call = parentcall) {
        e$call <- call
        stop(e)
      },
      warning = function(w, call = parentcall) {
        w$call <- call
        warning(w)
        invokeRestart("muffleWarning")
      },
      message = function(m, call = parentcall) {
        m$call <- call
      }
    )
  )
}

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

#' Encode Character Object as for UTF-8 Printing
#'
#' Escape the strings in a character object, optionally adding quotes or
#' spaces, adjusting the width for display.
#'
#' `utf8_encode()` encodes a character object for printing on a UTF-8 device
#' by escaping controls characters and other non-printable characters. When
#' `display = TRUE`, the function optimizes the encoding for display by
#' removing default ignorable characters (soft hyphens, zero-width spaces,
#' etc.) and placing zero-width spaces after wide emoji. When
#' `output_utf8()` is `FALSE` the function escapes all non-ASCII
#' characters and gives the same results on all platforms.
#'
#' @inheritParams rlang::args_dots_empty
#' @param x character object.
#' @param width integer giving the minimum field width; specify `NULL` or
#'   `NA` for no minimum.
#' @param quote logical scalar indicating whether to surround results with
#'   double-quotes and escape internal double-quotes.
#' @param justify justification; one of `"left"`, `"right"`,
#'   `"centre"`, or `"none"`. Can be abbreviated.
#' @param escapes a character string specifying the display style for the
#'   backslash escapes, as an ANSI SGR parameter string, or NULL for no styling.
#' @param display logical scalar indicating whether to optimize the encoding
#'   for display, not byte-for-byte data transmission.
#' @param utf8 logical scalar indicating whether to encode for a UTF-8 capable
#'   display (ASCII-only otherwise), or `NULL` to encode for output
#'   capabilities as determined by `output_utf8()`.
#' @return A character object with the same attributes as `x` but with
#'   `Encoding` set to `"UTF-8"`.
#' @seealso [utf8_print()].
#' @examples
#'
#' # the second element is encoded in latin-1, but declared as UTF-8
#' x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
#' Encoding(x) <- c("UTF-8", "UTF-8", "bytes")
#'
#' # encoding
#' utf8_encode(x)
#'
#' # add style to the escapes
#' cat(utf8_encode("hello\nstyled\\world", escapes = "1"), "\n")
#'
#' @export utf8_encode
utf8_encode <- function(
  x,
  ...,
  width = 0L,
  quote = FALSE,
  justify = "left",
  escapes = NULL,
  display = FALSE,
  utf8 = NULL
) {
  stopifnot(...length() == 0)

  if (is.null(x)) {
    return(NULL)
  }

  if (!is.character(x)) {
    stop("argument is not a character object")
  }

  with_rethrow({
    width <- as_integer_scalar("width", width)
    quote <- as_option("quote", quote)
    justify <- as_justify("justify", justify)
    escapes <- as_style("escapes", escapes)
    display <- as_option("display", display)
    utf8 <- as_output_utf8("utf8", utf8)
  })

  .Call(rutf8_utf8_encode, x, width, quote, justify, escapes, display, utf8)
}

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

#' UTF-8 Text Formatting
#'
#' Format a character object for UTF-8 printing.
#'
#' `utf8_format()` formats a character object for printing, optionally
#' truncating long character strings.
#'
#' @inheritParams rlang::args_dots_empty
#' @param x character object.
#' @param trim logical scalar indicating whether to suppress padding spaces
#'   around elements.
#' @param chars integer scalar indicating the maximum number of character units
#'   to display.  Wide characters like emoji take two character units; combining
#'   marks and default ignorables take none. Longer strings get truncated and
#'   suffixed or prefixed with an ellipsis (`"..."` or `"\u2026"`,
#'   whichever is most appropriate for the current character locale). Set to
#'   `NULL` to limit output to the line width as determined by
#'   `getOption("width")`.
#' @param justify justification; one of `"left"`, `"right"`,
#'   `"centre"`, or `"none"`. Can be abbreviated.
#' @param width the minimum field width; set to `NULL` or `0` for no
#'   restriction.
#' @param na.encode logical scalar indicating whether to encode `NA`
#'   values as character strings.
#' @param quote logical scalar indicating whether to format for a context with
#'   surrounding double-quotes (`'"'`) and escaped internal double-quotes.
#' @param na.print character string (or `NULL`) indicating the encoding
#'   for `NA` values. Ignored when `na.encode` is `FALSE`.
#' @param print.gap non-negative integer (or `NULL`) giving the number of
#'   spaces in gaps between columns; set to `NULL` or `1` for a single
#'   space.
#' @param utf8 logical scalar indicating whether to format for a UTF-8 capable
#'   display (ASCII-only otherwise), or `NULL` to format for output
#'   capabilities as determined by `output_utf8()`.
#' @return A character object with the same attributes as `x` but with
#'   `Encoding` set to `"UTF-8"` for elements that can be converted to
#'   valid UTF-8 and `"bytes"` for others.
#' @seealso [utf8_print()], [utf8_encode()].
#' @examples
#'
#' # the second element is encoded in latin-1, but declared as UTF-8
#' x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
#' Encoding(x) <- c("UTF-8", "UTF-8", "bytes")
#'
#' # formatting
#' utf8_format(x, chars = 3)
#' utf8_format(x, chars = 3, justify = "centre", width = 10)
#' utf8_format(x, chars = 3, justify = "right")
#'
#' @export utf8_format
utf8_format <- function(
  x,
  ...,
  trim = FALSE,
  chars = NULL,
  justify = "left",
  width = NULL,
  na.encode = TRUE,
  quote = FALSE,
  na.print = NULL,
  print.gap = NULL,
  utf8 = NULL
) {
  stopifnot(...length() == 0)

  if (is.null(x)) {
    return(NULL)
  }

  if (!is.character(x)) {
    stop("argument is not a character object")
  }

  with_rethrow({
    trim <- as_option("trim", trim)
    chars <- as_chars("chars", chars)
    justify <- as_justify("justify", justify)
    width <- as_integer_scalar("width", width)
    na.encode <- as_option("na.encode", na.encode)
    quote <- as_option("quote", quote)
    na.print <- as_na_print("na.print", na.print)
    print.gap <- as_print_gap("print_gap", print.gap)
    utf8 <- as_output_utf8("utf8", utf8)
  })

  ellipsis <- "\u2026"
  iellipsis <- iconv(ellipsis, "UTF-8", "")
  if (is.na(iellipsis) || identical(iellipsis, "...")) {
    ellipsis <- "..."
    wellipsis <- 3L
  } else {
    wellipsis <- 1L
  }

  if (is.null(chars) && length(x) > 0) {
    linewidth <- getOption("width")
    quotes <- if (quote) 2 else 0
    gap <- if (is.null(print.gap)) 1 else NULL

    dim <- dim(x)
    dimnames <- dimnames(x)
    names <- if (is.null(dimnames)) names(x) else dimnames[[1]]

    if (is.null(names)) {
      comma <- length(dim) > 1
      namewidth <- floor(log10(length(x)) + 1) + 2 + comma
    } else if (length(dim) > 1) {
      namewidth <- max(0, utf8_width(names, utf8 = utf8))
    } else {
      namewidth <- 0
    }

    chars <- (linewidth - wellipsis - quotes - gap - namewidth)
    chars <- max(chars, 12)
  }

  .Call(
    rutf8_utf8_format,
    x,
    trim,
    chars,
    justify,
    width,
    na.encode,
    quote,
    na.print,
    ellipsis,
    wellipsis,
    utf8
  )
}

if (utils::packageVersion("testthat") <= "1.0.2") {
    capture_output <- function(code, print = FALSE, width = 80) {
        oldwidth <- getOption("width")
        if (width != oldwidth) {
            skip_if_not(with(R.Version(), paste(major, minor, sep = "."))
                        >= "3.4.0", "Setting width fails on R < 3.4.0")
            
            options(width = width)
            on.exit(options(width = oldwidth), add = TRUE)
        }
        testthat::capture_output(code, print)
    }
}

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# converts a character vector from its declared encoding to UTF-8

#' UTF-8 Character Encoding
#'
#' @description
#' UTF-8 text encoding and validation
#'
#' `as_utf8()` converts a character object from its declared encoding to a
#' valid UTF-8 character object, or throws an error if no conversion is
#' possible. If `normalize = TRUE`, then the text gets transformed to
#' Unicode composed normal form (NFC) after conversion to UTF-8.
#'
#' `utf8_valid()` tests whether the elements of a character object can be
#' translated to valid UTF-8 strings.
#'
#' @aliases as_utf8 utf8_valid
#' @param x character object.
#' @param normalize a logical value indicating whether to convert to Unicode
#'   composed normal form (NFC).
#' @return For `as_utf8()`, the result is a character object with the same
#'   attributes as `x` but with `Encoding` set to `"UTF-8"`.
#'
#'   For `utf8_valid()` a logical object with the same `names`,
#'   `dim`, and `dimnames` as `x`.
#' @seealso [utf8_normalize()], [iconv()].
#' @examples
#'
#' # the second element is encoded in latin-1, but declared as UTF-8
#' x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
#' Encoding(x) <- c("UTF-8", "UTF-8", "bytes")
#'
#' # attempt to convert to UTF-8 (fails)
#' \dontrun{as_utf8(x)}
#'
#' y <- x
#' Encoding(y[2]) <- "latin1" # mark the correct encoding
#' as_utf8(y) # succeeds
#'
#' # test for valid UTF-8
#' utf8_valid(x)
#'
#' @export as_utf8
as_utf8 <- function(x, normalize = FALSE) {
  ans <- .Call(rutf8_as_utf8, x)
  if (normalize) {
    ans <- utf8_normalize(ans)
  }
  ans
}

# test whether the elements can be converted to valid UTF-8
#' @rdname as_utf8
#' @export
utf8_valid <- function(x) {
  .Call(rutf8_utf8_valid, x)
}

# gets the width; NA for invalid or nonprintable sequences

#' Measure the Character String Width
#'
#' Compute the display widths of the elements of a character object.
#'
#' `utf8_width()` returns the printed widths of the elements of a character
#' object on a UTF-8 device (or on an ASCII device when `output_utf8()` is
#' `FALSE`), when printed with `utf8_print()`. If the string is not
#' printable on the device, for example if it contains a control code like
#' `"\n"`, then the result is `NA`. If `encode = TRUE`, the
#' default, then the function returns the widths of the encoded elements via
#' `utf8_encode()`); otherwise, the function returns the widths of the
#' original elements.
#'
#' @inheritParams rlang::args_dots_empty
#' @param x character object.
#' @param encode whether to encode the object before measuring its width.
#' @param quote whether to quote the object before measuring its width.
#' @param utf8 logical scalar indicating whether to determine widths assuming a
#'   UTF-8 capable display (ASCII-only otherwise), or `NULL` to format for
#'   output capabilities as determined by `output_utf8()`.
#' @return An integer object, with the same `names`, `dim`, and
#'   `dimnames` as `x`.
#' @seealso [utf8_print()].
#' @examples
#'
#' # the second element is encoded in latin-1, but declared as UTF-8
#' x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
#' Encoding(x) <- c("UTF-8", "UTF-8", "bytes")
#'
#' # get widths
#' utf8_width(x)
#' utf8_width(x, encode = FALSE)
#' utf8_width('"')
#' utf8_width('"', quote = TRUE)
#'
#' @export utf8_width
utf8_width <- function(
  x,
  ...,
  encode = TRUE,
  quote = FALSE,
  utf8 = NULL
) {
  stopifnot(...length() == 0)

  with_rethrow({
    encode <- as_option("encode", encode)
    quote <- as_option("quote", quote)
    utf8 <- as_output_utf8("utf8", utf8)
  })

  .Call(rutf8_utf8_width, x, encode, quote, utf8)
}


#' Text Normalization
#'
#' Transform text to normalized form, optionally mapping to lowercase and
#' applying compatibility maps.
#'
#' `utf8_normalize()` converts the elements of a character object to Unicode
#' normalized composed form (NFC) while applying the character maps specified
#' by the `map_case`, `map_compat`, `map_quote`, and
#' `remove_ignorable` arguments.
#'
#' @inheritParams rlang::args_dots_empty
#' @param x character object.
#' @param map_case a logical value indicating whether to apply Unicode case
#'   mapping to the text. For most languages, this transformation changes
#'   uppercase characters to their lowercase equivalents.
#' @param map_compat a logical value indicating whether to apply Unicode
#'   compatibility mappings to the characters, those required for NFKC and NFKD
#'   normal forms.
#' @param map_quote a logical value indicating whether to replace curly single
#'   quotes and Unicode apostrophe characters with ASCII apostrophe (U+0027).
#' @param remove_ignorable a logical value indicating whether to remove Unicode
#'   "default ignorable" characters like zero-width spaces and soft hyphens.
#' @return The result is a character object with the same attributes as
#'   `x` but with `Encoding` set to `"UTF-8"`.
#' @seealso [as_utf8()].
#' @examples
#'
#' angstrom <- c("\u00c5", "\u0041\u030a", "\u212b")
#' utf8_normalize(angstrom) == "\u00c5"
#'
#' @export utf8_normalize
utf8_normalize <- function(
  x,
  ...,
  map_case = FALSE,
  map_compat = FALSE,
  map_quote = FALSE,
  remove_ignorable = FALSE
) {
  stopifnot(...length() == 0)

  with_rethrow({
    x <- as_utf8(x, normalize = FALSE)
    map_case <- as_option("map_case", map_case)
    map_compat <- as_option("map_compat", map_compat)
    map_quote <- as_option("map_quote", map_quote)
    remove_ignorable <- as_option("remove_ignorable", remove_ignorable)
  })

  .Call(
    rutf8_utf8_normalize,
    x,
    map_case,
    map_compat,
    map_quote,
    remove_ignorable
  )
}

# From http://unicode.org/reports/tr15/
test_that("'utf8_normalize' can reproduce Fig. 3", {
  src <- c("\u212b", "\u2126")
  nfd <- c("\u0041\u030a", "\u03a9")
  nfc <- c("\u00c5", "\u03a9")

  expect_equal(utf8_normalize(src), nfc)
  expect_equal(utf8_normalize(nfd), nfc)
})

# From http://unicode.org/reports/tr15/
test_that("'utf8_normalize' can reproduce Fig. 4", {
  src <- c("\u00c5", "\u00f4")
  nfd <- c("\u0041\u030a", "\u006f\u0302")
  nfc <- c("\u00c5", "\u00f4")

  expect_equal(utf8_normalize(src), nfc)
  expect_equal(utf8_normalize(nfd), nfc)
})

# From http://unicode.org/reports/tr15/
test_that("'utf8_normalize' can reproduce Fig. 5", {
  src <- c("\u1e69", "\u1e0b\u0323", "\u0071\u0307\u0323")
  nfd <- c("\u0073\u0323\u0307", "\u0064\u0323\u0307", "\u0071\u0323\u0307")
  nfc <- c("\u1e69", "\u1e0d\u0307", "\u0071\u0323\u0307")

  expect_equal(utf8_normalize(src), nfc)
  expect_equal(utf8_normalize(nfd), nfc)
})

# From http://unicode.org/reports/tr15/
test_that("'utf8_normalize' can reproduce Fig. 6", {
  src <- c("\ufb01", "\u0032\u2075", "\u1e9b\u0323")
  nfd <- c("\ufb01", "\u0032\u2075", "\u017f\u0323\u0307")
  nfc <- c("\ufb01", "\u0032\u2075", "\u1e9b\u0323")
  nfkd <- c("\u0066\u0069", "\u0032\u0035", "\u0073\u0323\u0307")
  nfkc <- c("\u0066\u0069", "\u0032\u0035", "\u1e69")

  expect_equal(utf8_normalize(src), nfc)
  expect_equal(utf8_normalize(nfd), nfc)
  expect_equal(utf8_normalize(src, map_compat = TRUE), nfkc)
  expect_equal(utf8_normalize(nfd, map_compat = TRUE), nfkc)
  expect_equal(utf8_normalize(nfkd), nfkc)
})


test_that("'utf8_normalize' can normalize, case fold, and remove ignorables", {
  src <- c("A", "\u00df", "\u1e9e", "\u1fc3", "\u200b")
  nfkc_casefold <- c("a", "ss", "ss", "\u03b7\u03b9", "")
  expect_equal(
    utf8_normalize(src, map_case = TRUE, remove_ignorable = TRUE),
    nfkc_casefold
  )
})


test_that("'utf8_normalize' can map quotes", {
  src <- c("\"", "'", "\u2018", "\u2019", "\u201c", "\u201d")
  quotefold <- c("\"", "'", "'", "'", "\u201c", "\u201d")
  expect_equal(utf8_normalize(src, map_quote = TRUE), quotefold)
})


test_that("'utf8_normalize' accepts NULL", {
  expect_equal(utf8_normalize(NULL), NULL)
})


test_that("'utf8_normalize' accepts NA", {
  expect_equal(utf8_normalize(NA_character_), NA_character_)
})


test_that("'utf8_normalize' can handle backslash", {
  expect_equal(utf8_normalize("\\m"), "\\m")
})

test_that("'as_utf8' errors on latin1 declared to be UTF-8", {
  x <- c("a", "b", "the command of her beauty, and her \xa320,000", "d")
  Encoding(x) <- "UTF-8"

  expect_equal(utf8_valid(x), c(TRUE, TRUE, FALSE, TRUE))
  expect_error(as_utf8(x), "entry 3 has wrong Encoding; marked as \"UTF-8\" but invalid leading byte (0xA3) at position 36", fixed = TRUE)
})


test_that("utf8_valid errors on invalid UTF-8", {
  x <- c("a", "b", "c", "d", "\xf8\x88\x80\x80\x80") # intToUtf8(0x00200000)
  Encoding(x) <- "UTF-8"

  expect_equal(utf8_valid(x), c(TRUE, TRUE, TRUE, TRUE, FALSE))
  expect_error(as_utf8(x), "entry 5 has wrong Encoding; marked as \"UTF-8\" but invalid leading byte (0xF8) at position 1", fixed = TRUE)
})


test_that("utf8_valid passes on valid UTF-8 in bytes encoding", {
  x <- "hello\u2002"
  Encoding(x) <- "bytes"
  expect_equal(utf8_valid(x), TRUE)

  y <- x
  Encoding(y) <- "UTF-8"
  expect_equal(as_utf8(x), enc2utf8(y))
})


test_that("utf8_valid passes on valid ASCII in unknown encoding", {
  x <- "world"
  expect_equal(utf8_valid(x), TRUE)
  expect_equal(as_utf8(x), enc2utf8(x))
})


test_that("utf8_valid errors on invalid UTF8 in bytes encoding", {
  x <- paste0("hello", "\xfc\x8f\xbf\xbf\xbf\xbf") # intToUtf8(0xfffffff)
  Encoding(x) <- "bytes"
  expect_equal(utf8_valid(x), FALSE)
  expect_error(as_utf8(x), "entry 1 cannot be converted from \"bytes\" Encoding to \"UTF-8\"; invalid leading byte (0xFC) at position 6", fixed = TRUE)
})

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

#' @rdname output_utf8
#' @export
output_ansi <- function() {
  # tty?
  if (isatty(stdout())) {
    # and not Windows GUI?
    return(.Platform$GUI != "Rgui")
  }

  # RStudio 1.1 or later with color enabled?
  if (!is.na(as.numeric(Sys.getenv("RSTUDIO_CONSOLE_COLOR")))) {
    # and output is stdout?
    return(stdout() == 1)
  }

  FALSE
}


#' Output Capabilities
#'
#' Test whether the output connection has ANSI style escape support or UTF-8
#' support.
#'
#' `output_ansi()` tests whether the output connection supports ANSI style
#' escapes. This is `TRUE` if the connection is a terminal and not the
#' Windows GUI. Otherwise, it is true if running in RStudio 1.1 or later with
#' ANSI escapes enabled, provided `stdout()` has not been redirected to
#' another connection by `sink()`.
#'
#' `output_utf8()` tests whether the output connection supports UTF-8. For
#' most platforms `l10n_info()$"UTF-8"` gives this information, but this
#' does not give an accurate result for Windows GUIs. To work around this, we
#' proceed as follows: \itemize{
#'
#' \item if the character locale
#' (`LC_CTYPE`) is `"C"`, then the result is `FALSE`;
#'
#' \item otherwise, if `l10n_info()$"UTF-8"` is `TRUE`, then the
#' result is `TRUE`;
#'
#' \item if running on Windows, then the result is `TRUE`;
#'
#' \item in all other cases the result is `FALSE`. }
#'
#' Strictly speaking,
#' UTF-8 support is always available on Windows GUI, but only a subset of UTF-8
#' is available (defined by the current character locale) when the output is
#' redirected by `knitr` or another process. Unfortunately, it is
#' impossible to set the character locale to UTF-8 on Windows. Further, the
#' `utf8` package only handles two character locales: C and UTF-8.  To get
#' around this, on Windows, we treat all non-C locales on that platform as
#' UTF-8. This liberal approach means that characters in the user's locale
#' never get escaped; others will get output as `<U+XXXX>`, with incorrect
#' values for `utf8_width()`.
#'
#' @aliases output_ansi output_utf8
#' @return A logical scalar indicating whether the output connection supports
#'   the given capability.
#' @seealso [.Platform()], [isatty()],
#' [l10n_info()], [Sys.getlocale()]
#' @examples
#'
#' # test whether ANSI style escapes or UTF-8 output are supported
#' cat("ANSI:", output_ansi(), "\n")
#' cat("UTF8:", output_utf8(), "\n")
#'
#' # switch to C locale
#' Sys.setlocale("LC_CTYPE", "C")
#' cat("ANSI:", output_ansi(), "\n")
#' cat("UTF8:", output_utf8(), "\n")
#'
#' # switch to native locale
#' Sys.setlocale("LC_CTYPE", "")
#'
#' tmp <- tempfile()
#' sink(tmp) # redirect output to a file
#' cat("ANSI:", output_ansi(), "\n")
#' cat("UTF8:", output_utf8(), "\n")
#' sink() # restore stdout
#'
#' # inspect the output
#' readLines(tmp)
#'
#' @export output_utf8
output_utf8 <- function() {
  # ASCII-only character locale?
  if (Sys.getlocale("LC_CTYPE") == "C") {
    return(FALSE)
  }

  # UTF-8 locale?
  if (l10n_info()$`UTF-8`) {
    return(TRUE)
  }

  # Windows?
  if (.Platform$OS.type == "windows") {
    # This isn't really the case, but there's no way to set the
    # locale to UTF-8 on Windows. In RGui and RStudio, UTF-8 is
    # always supported on stdout(); output through connections
    # gets translated through the native locale.
    return(TRUE)
  }

  FALSE
}

#' The utf8 Package
#'
#' UTF-8 Text Processing
#'
#' Functions for manipulating and printing UTF-8 encoded text:
#'
#' \itemize{ \item [as_utf8()] attempts to convert character data to
#' UTF-8, throwing an error if the data is invalid;
#'
#' \item [utf8_valid()] tests whether character data is valid
#' according to its declared encoding;
#'
#' \item [utf8_normalize()] converts text to Unicode composed normal
#' form (NFC), optionally applying case-folding and compatibility maps;
#'
#' \item [utf8_encode()] encodes a character string, escaping all
#' control characters, so that it can be safely printed to the screen;
#'
#' \item [utf8_format()] formats a character vector by truncating to
#' a specified character width limit or by left, right, or center justifying;
#'
#' \item [utf8_print()] prints UTF-8 character data to the screen;
#'
#' \item [utf8_width()] measures the display width of UTF-8 character
#' strings (many emoji and East Asian characters are twice as wide as other
#' characters);
#'
#' \item [output_ansi()] and [output_utf8()] test for the
#' output connections capabilities. }
#'
#' For a complete list of functions, use `library(help = "utf8")`.
#'
#' @useDynLib utf8, .registration = TRUE
"_PACKAGE"

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

#' Print UTF-8 Text
#'
#' Print a UTF-8 character object.
#'
#' `utf8_print()` prints a character object after formatting it with
#' [utf8_format()].
#'
#' For ANSI terminal output (when `output_ansi()` is `TRUE`), you can
#' style the row and column names with the `rownames` and `names`
#' parameters, specifying an ANSI SGR parameter string; see
#' <https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters>.
#'
#' @inheritParams rlang::args_dots_empty
#' @param x character object.
#' @param chars integer scalar indicating the maximum number of character units
#'   to display.  Wide characters like emoji take two character units; combining
#'   marks and default ignorables take none. Longer strings get truncated and
#'   suffixed or prefixed with an ellipsis (`"..."` in C locale,
#'   `"\u2026"` in others). Set to `NULL` to limit output to the line
#'   width as determined by `getOption("width")`.
#' @param quote logical scalar indicating whether to put surrounding
#'   double-quotes (`'"'`) around character strings and escape internal
#'   double-quotes.
#' @param na.print character string (or `NULL`) indicating the encoding
#'   for `NA` values. Ignored when `na.encode` is `FALSE`.
#' @param print.gap non-negative integer (or `NULL`) giving the number of
#'   spaces in gaps between columns; set to `NULL` or `1` for a single
#'   space.
#' @param right logical scalar indicating whether to right-justify character
#'   strings.
#' @param max non-negative integer (or `NULL`) indicating the maximum
#'   number of elements to print; set to `getOption("max.print")` if
#'   argument is `NULL`.
#' @param names a character string specifying the display style for the
#'   (column) names, as an ANSI SGR parameter string.
#' @param rownames a character string specifying the display style for the row
#'   names, as an ANSI SGR parameter string.
#' @param escapes a character string specifying the display style for the
#'   backslash escapes, as an ANSI SGR parameter string.
#' @param display logical scalar indicating whether to optimize the encoding
#'   for display, not byte-for-byte data transmission.
#' @param style logical scalar indicating whether to apply ANSI terminal escape
#'   codes to style the output.  Ignored when `output_ansi()` is
#'   `FALSE`.
#' @param utf8 logical scalar indicating whether to optimize results for a
#'   UTF-8 capable display, or `NULL` to set as the result of
#'   `output_utf8()`. Ignored when `output_utf8()` is `FALSE`.
#' @return The function returns `x` invisibly.
#' @seealso [utf8_format()].
#' @examples
#'
#' # printing (assumes that output is capable of displaying Unicode 10.0.0)
#' print(intToUtf8(0x1F600 + 0:79)) # with default R print function
#' utf8_print(intToUtf8(0x1F600 + 0:79)) # with utf8_print, truncates line
#' utf8_print(intToUtf8(0x1F600 + 0:79), chars = 1000) # higher character limit
#'
#' # in C locale, output ASCII (same results on all platforms)
#' oldlocale <- Sys.getlocale("LC_CTYPE")
#' invisible(Sys.setlocale("LC_CTYPE", "C")) # switch to C locale
#' utf8_print(intToUtf8(0x1F600 + 0:79))
#' invisible(Sys.setlocale("LC_CTYPE", oldlocale)) # switch back to old locale
#'
#' # Mac and Linux only: style the names
#' # see https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
#' utf8_print(matrix(as.character(1:20), 4, 5),
#'            names = "1;4", rownames = "2;3")
#'
#' @export utf8_print
utf8_print <- function(
  x,
  ...,
  chars = NULL,
  quote = TRUE,
  na.print = NULL,
  print.gap = NULL,
  right = FALSE,
  max = NULL,
  names = NULL,
  rownames = NULL,
  escapes = NULL,
  display = TRUE,
  style = TRUE,
  utf8 = NULL
) {
  stopifnot(...length() == 0)

  if (is.null(x)) {
    return(invisible(NULL))
  }

  if (!is.character(x)) {
    stop("argument is not a character object")
  }

  with_rethrow({
    chars <- as_chars("chars", chars)
    quote <- as_option("quote", quote)
    na.print <- as_na_print("na.print", na.print)
    print.gap <- as_print_gap("print_gap", print.gap)
    right <- as_option("right", right)
    max <- as_max_print("max", max)
    names <- as_style("names", names)
    rownames <- as_style("rownames", rownames)
    escapes <- as_style("escapes", escapes)
    display <- as_option("display", display)
    style <- as_option("style", style)
    utf8 <- as_output_utf8("utf8", utf8)
  })

  if (is.null(print.gap)) {
    print.gap <- 1L
  }
  if (is.null(max)) {
    max <- getOption("max.print")
  }

  if (!output_ansi()) {
    style <- FALSE
  }

  if (!output_utf8()) {
    utf8 <- FALSE
  }

  # truncate character objects
  justify <- if (right) "right" else "left"
  fmt <- utf8_format(
    x,
    trim = TRUE,
    chars = chars,
    justify = justify,
    na.encode = FALSE,
    quote = quote,
    utf8 = utf8
  )

  dim <- dim(x)
  if (is.null(dim) || length(dim) == 1) {
    nprint <- print_vector(
      fmt,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
  } else if (length(dim) == 2) {
    nprint <- print_matrix(
      fmt,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
  } else {
    nprint <- print_array(
      fmt,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
  }

  n <- length(x)
  if (nprint < n) {
    cat(sprintf(
      " [ reached getOption(\"max.print\") -- omitted %d entries ]\n",
      n - nprint
    ))
  }

  invisible(x)
}


print_vector <- function(
  x,
  quote,
  na.print,
  print.gap,
  right,
  max,
  names,
  rownames,
  escapes,
  display,
  style,
  utf8
) {
  if (length(x) == 0) {
    cat("character(0)\n")
    return(0L)
  }

  # drop dim, convert dimnames to names
  if (!is.null(dim(x))) {
    x <- c(x)
  }

  if (!is.null(names(x))) {
    nprint <- print_vector_named(
      x,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
  } else {
    nprint <- print_vector_unnamed(
      x,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
  }

  nprint
}


print_vector_named <- function(
  x,
  quote,
  na.print,
  print.gap,
  right,
  max,
  names,
  rownames,
  escapes,
  display,
  style,
  utf8
) {
  n <- length(x)
  nm <- names(x)
  namewidth <- max(0L, utf8_width(nm, utf8 = utf8))
  eltwidth <- element_width(x, quote = quote, na.print = na.print, utf8 = utf8)
  width <- max(eltwidth, namewidth)

  linewidth <- getOption("width")
  ncol <- max(1L, linewidth %/% (width + print.gap))
  extra <- n %% ncol

  nprint <- 0L
  off <- 0L
  while (off + ncol <= n && nprint < max) {
    ix <- (off + 1L):(off + ncol)
    mat <- matrix(
      x[ix],
      ncol = ncol,
      byrow = TRUE,
      dimnames = list(NULL, nm[ix])
    )
    np <- print_table(
      mat,
      width = width,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max - nprint,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
    nprint <- nprint + np
    off <- off + ncol
  }

  if (extra > 0L && nprint < max) {
    ix <- n - extra + seq_len(extra)
    last <- rbind(as.vector(x[ix]))
    rownames(last) <- NULL
    colnames(last) <- nm[ix]

    np <- print_table(
      last,
      width = width,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max - nprint,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
    nprint <- nprint + np
  }

  nprint
}


print_vector_unnamed <- function(
  x,
  quote,
  na.print,
  print.gap,
  right,
  max,
  names,
  rownames,
  escapes = escapes,
  display,
  style,
  utf8
) {
  n <- length(x)
  nm <- utf8_format(
    paste0("[", seq_len(n), "]"),
    justify = "right",
    utf8 = utf8
  )
  namewidth <- max(0L, utf8_width(nm, utf8 = utf8))
  width <- element_width(x, quote = quote, na.print = na.print, utf8 = utf8)

  linewidth <- getOption("width")
  ncol <- max(1L, (linewidth - namewidth) %/% (width + print.gap))
  extra <- n %% ncol

  mat <- matrix(x[seq_len(n - extra)], ncol = ncol, byrow = TRUE)
  rownames(mat) <- nm[seq(from = 1L, by = ncol, length.out = nrow(mat))]

  nprint <- print_table(
    mat,
    width = width,
    quote = quote,
    na.print = na.print,
    print.gap = print.gap,
    right = right,
    max = max,
    names = names,
    rownames = rownames,
    escapes = escapes,
    display = display,
    style = style,
    utf8 = utf8
  )

  if (extra > 0L && nprint < max) {
    last <- rbind(as.vector(x[n - extra + seq_len(extra)]))
    rownames(last) <- nm[n - extra + 1]
    np <- print_table(
      last,
      width = width,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max - nprint,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
    nprint <- nprint + np
  }

  nprint
}


element_width <- function(x, quote, na.print, utf8) {
  width <- max(
    0L,
    utf8_width(x, encode = TRUE, quote = quote, utf8 = utf8),
    na.rm = TRUE
  )
  if (anyNA(x)) {
    if (is.null(na.print)) {
      na.print <- if (quote) "NA" else "<NA>"
    }
    width <- max(width, utf8_width(na.print, utf8 = utf8))
  }
  width
}


print_matrix <- function(
  x,
  quote,
  na.print,
  print.gap,
  right,
  max,
  names,
  rownames,
  escapes,
  display,
  style,
  utf8
) {
  if (all(dim(x) == 0)) {
    cat("<0 x 0 matrix>\n")
    return(0L)
  }
  x <- set_dimnames(x)
  print_table(
    x,
    width = 0L,
    quote = quote,
    na.print = na.print,
    print.gap = print.gap,
    right = right,
    max = max,
    names = names,
    rownames = rownames,
    escapes = escapes,
    display = display,
    style = style,
    utf8 = utf8
  )
}


print_array <- function(
  x,
  quote,
  na.print,
  print.gap,
  right,
  max,
  names,
  rownames,
  escapes,
  display,
  style,
  utf8
) {
  n <- length(x)
  dim <- dim(x)
  if (any(dim == 0)) {
    cat(sprintf("<%s array>\n", paste(dim, collapse = " x ")))
    return(0L)
  }

  x <- set_dimnames(x)
  dimnames <- dimnames(x)

  nrow <- dim[1]
  ncol <- dim[2]

  base <- c(NA, NA, rep(1, length(dim) - 2))
  label <- vector("character", length(dim))
  for (r in 3:length(dim)) {
    label[[r]] <- dimnames[[r]][[1]]
  }

  nprint <- 0L
  off <- 0L
  while (off + nrow * ncol <= n && nprint < max) {
    cat(paste(label, collapse = ", "), "\n\n", sep = "")

    ix <- off + seq_len(nrow * ncol)
    mat <- matrix(x[ix], nrow, ncol, dimnames = dimnames[1:2])
    np <- print_table(
      mat,
      width = 0L,
      quote = quote,
      na.print = na.print,
      print.gap = print.gap,
      right = right,
      max = max - nprint,
      names = names,
      rownames = rownames,
      escapes = escapes,
      display = display,
      style = style,
      utf8 = utf8
    )
    nprint <- nprint + np
    off <- off + (nrow * ncol)

    r <- 3L
    while (r < length(dim) && base[r] == dim[r]) {
      base[r] <- 1L
      label[r] <- dimnames[[r]][[1L]]
      r <- r + 1L
    }
    if (base[r] < dim[r]) {
      base[r] <- base[r] + 1L
      label[r] <- dimnames[[r]][[base[r]]]
    }
    cat("\n")
  }

  nprint
}


print_table <- function(
  x,
  width,
  quote,
  na.print,
  print.gap,
  right,
  max,
  names,
  rownames,
  escapes,
  display,
  style,
  utf8
) {
  width <- as.integer(width)
  if (is.null(na.print)) {
    na.print <- if (quote) "NA" else "<NA>"
    na.name.print <- "<NA>"
  } else {
    na.name.print <- na.print
  }
  print.gap <- as.integer(print.gap)
  max <- as.integer(max)

  if (!is.null(rownames(x))) {
    rownames(x)[is.na(rownames(x))] <- na.name.print
  }
  if (!is.null(colnames(x))) {
    colnames(x)[is.na(colnames(x))] <- na.name.print
  }

  linewidth <- getOption("width")
  str <- .Call(
    rutf8_render_table,
    x,
    width,
    quote,
    na.print,
    print.gap,
    right,
    max,
    names,
    rownames,
    escapes,
    display,
    style,
    utf8,
    linewidth
  )
  cat(str)

  nprint <- min(max, length(x))
  nprint
}


set_dimnames <- function(x) {
  dim <- dim(x)
  dimnames <- dimnames(x)

  if (is.null(dimnames)) {
    dimnames <- vector("list", length(dim))
  }

  for (i in seq_along(dim)) {
    d <- dim[[i]]
    ix <- seq_len(d)
    if (is.null(dimnames[[i]]) && d > 0) {
      if (i == 1) {
        dimnames[[i]] <- format(paste0("[", ix, ",]"), justify = "right")
      } else if (i == 2) {
        dimnames[[i]] <- paste0("[,", ix, "]")
      } else {
        dimnames[[i]] <- as.character(ix)
      }
    }
  }

  dimnames(x) <- dimnames

  x
}

#  Copyright 2017 Patrick O. Perry.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


as_character_scalar <- function(name, value, utf8 = TRUE) {
  if (is.null(value)) {
    return(NULL)
  }
  value <- as_character_vector(name, value, utf8)
  if (length(value) != 1) {
    stop(sprintf("'%s' must be a scalar character string", name))
  }
  value
}


as_character_vector <- function(name, value, utf8 = TRUE) {
  if (!(is.null(value) || is.character(value) || is.factor(value) ||
    all(is.na(value)))) {
    stop(sprintf("'%s' must be text, a character vector, or NULL", name))
  }
  if (is.null(value)) {
    return(NULL)
  }
  value <- as.character(value)
  if (utf8) {
    value <- as_utf8(value)
  }
  value
}


as_enum <- function(name, value, choices) {
  if (!(is.character(value) && length(value) == 1 && !is.na(value))) {
    stop(sprintf("'%s' must be a character string", name))
  }

  i <- pmatch(value, choices, nomatch = 0)
  if (all(i == 0)) {
    stop(
      sprintf("'%s' must be one of the following: ", name),
      paste(dQuote(choices), collapse = ", ")
    )
  }
  i <- i[i > 0]
  choices[[i]]
}


as_integer_scalar <- function(name, value, nonnegative = FALSE) {
  if (is.null(value)) {
    return(NULL)
  }
  value <- as_integer_vector(name, value, nonnegative)
  if (length(value) != 1) {
    stop(sprintf("'%s' must have length 1", name))
  }
  value
}


as_integer_vector <- function(name, value, nonnegative = FALSE) {
  if (is.null(value)) {
    return(NULL)
  }

  if (!(is.numeric(value) || all(is.na(value)))) {
    stop(sprintf("'%s' must be integer-valued", name))
  }

  value <- as.integer(value)
  if (nonnegative && any(!is.na(value) & value < 0)) {
    stop(sprintf("'%s' must be non-negative", name))
  }

  value
}


as_na_print <- function(name, value) {
  if (is.null(value)) {
    return(NULL)
  }
  value <- as_character_scalar(name, value)
  if (is.na(value)) {
    stop(sprintf("'%s' cannot be NA", name))
  }
  value
}


as_nonnegative <- function(name, value) {
  if (is.null(value)) {
    return(NULL)
  }
  value <- as_integer_scalar(name, value, nonnegative = TRUE)
  if (is.na(value)) {
    stop(sprintf("'%s' cannot be NA", name))
  }
  value
}


as_option <- function(name, value) {
  if (is.null(value)) {
    return(FALSE)
  }

  if (!(length(value) == 1 && is.logical(value) && !is.na(value))) {
    stop(sprintf("'%s' must be TRUE or FALSE", name))
  }
  as.logical(value)
}

as_chars <- as_nonnegative

as_justify <- function(name, value) {
  as_enum(name, value, c("left", "right", "centre", "none"))
}

as_max_print <- as_nonnegative

as_print_gap <- function(name, value) {
  value <- as_nonnegative(name, value)
  if (!is.null(value) && value > 1024) {
    stop(sprintf("'%s' must be less than or equal to 1024", name))
  }
  value
}

as_style <- function(name, value) {
  value <- as_character_scalar(name, value)
  if (is.null(value) || is.na(value)) {
    return(NULL)
  }
  if (!grepl("^[0-9;]*$", value)) {
    stop(sprintf("'%s' must be a valid ANSI SGR parameter string", name))
  }
  if (nchar(value) >= 128) {
    stop(sprintf("'%s' must have length below 128 characters", name))
  }
  value
}

as_output_utf8 <- function(name, value) {
  if (is.null(value)) {
    return(output_utf8())
  }
  as_option(name, value)
}

---
title: "Unicode: Emoji, accents, and international text"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Unicode: Emoji, accents, and international text}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



Character encoding
------------------

Before we can analyze a text in R, we first need to get its digital
representation, a sequence of ones and zeros. In practice this works by first
choosing an *encoding* for the text that assigns each character a numerical
value, and then translating the sequence of characters in the text to the
corresponding sequence of numbers specified by the encoding. Today, most new
text is encoded according to the [Unicode standard][unicode], specifically the
8-bit block Unicode Transfer Format, [UTF-8][utf8].  Joel Spolsky gives a
good overview of the situation in an [essay from 2003][spolsky2003].


The software community has mostly moved to UTF-8 as a standard for text
storage and interchange, but there is still a large volume of text in other
encodings. Whenever you read a text file into R, you need to specify the
encoding. If you don't, R will try to guess the encoding, and if it guesses
incorrectly, it will wrongly interpret the sequence of ones and zeros.


We will demonstrate the difficulties of encodings with the text of
Jane Austen's novel, _Mansfield Park_ provided by
[Project Gutenberg][gutenberg]. We will download the text, then
read in the lines of the novel.


```r
# download the zipped text from a Project Gutenberg mirror
url <-  "http://mirror.csclub.uwaterloo.ca/gutenberg/1/4/141/141.zip"
tmp <- tempfile()
download.file(url, tmp)

# read the text from the zip file
con <- unz(tmp, "141.txt", encoding = "UTF-8")
lines <- readLines(con)
close(con)
```

The `unz` function and other similar file connection functions have `encoding`
arguments which, if left unspecified default to assuming that text is encoded
in your operating system's native encoding. To ensure consistent behavior
across all platforms (Mac, Windows, and Linux), you should set this option
explicitly. Here, we set `encoding = "UTF-8"`. This is a reasonable default,
but it is not always appropriate. In general, you should determine the
appropriate `encoding` value by looking at the file. Unfortunately, the file
extension `".txt"` is not informative, and could correspond to any encoding.
However, if we read the first few lines of the file, we see the following:


```r
lines[11:20]
```

```
 [1] "Author: Jane Austen"                                         
 [2] ""                                                            
 [3] "Release Date: June, 1994  [Etext #141]"                      
 [4] "Posting Date: February 11, 2015"                             
 [5] ""                                                            
 [6] "Language: English"                                           
 [7] ""                                                            
 [8] "Character set encoding: ASCII"                               
 [9] ""                                                            
[10] "*** START OF THIS PROJECT GUTENBERG EBOOK MANSFIELD PARK ***"
```

The character set encoding is reported as ASCII, which is a subset of UTF-8.
So, we should be in good shape.


Unfortunately, we run into trouble as soon as we try to process the text:


```r
corpus::term_stats(lines) # produces an error
```

```
Error in corpus::term_stats(lines): argument entry 15252 is incorrectly marked as "UTF-8": invalid leading byte (0xA3) at position 36
```

The error message tells us that line 15252 contains an invalid byte.


```r
lines[15252]
```

```
[1] "the command of her beauty, and her \xa320,000, any one who could satisfy the"
```

We might wonder if there are other lines with invalid data.  We can find
all such lines using the `utf8_valid` function:

```r
lines[!utf8_valid(lines)]
```

```
[1] "the command of her beauty, and her \xa320,000, any one who could satisfy the"
```
So, there are no other invalid lines.

The offending byte in line 15252 is displayed as `\xa3`, an escape code
for hexadecimal value 0xa3, decimal value 163. To understand why this
is invalid, we need to learn more about UTF-8 encoding.


UTF-8
-----

### ASCII

The smallest unit of data transfer on modern computers is the byte, a sequence
of eight ones and zeros that can encode a number between 0 and 255
(hexadecimal 0x00 and 0xff). In the earliest character encodings, the numbers
from 0 to 127 (hexadecimal 0x00 to 0x7f) were standardized in an encoding
known as ASCII, the American Standard Code for Information Interchange.
Here are the characters corresponding to these codes:


```r
codes <- matrix(0:127, 8, 16, byrow = TRUE,
                dimnames = list(0:7, c(0:9, letters[1:6])))
ascii <- apply(codes, c(1, 2), intToUtf8)

# replace control codes with ""
ascii["0", c(0:6, "e", "f")] <- ""
ascii["1",] <- ""
ascii["7", "f"] <- ""

utf8_print(ascii, quote = FALSE)
```

```
  0 1 2 3 4 5 6 7  8  9  a  b  c  d  e f
0               \a \b \t \n \v \f \r    
1                                       
2   ! " # $ % & '  (  )  *  +  ,  -  . /
3 0 1 2 3 4 5 6 7  8  9  :  ;  <  =  > ?
4 @ A B C D E F G  H  I  J  K  L  M  N O
5 P Q R S T U V W  X  Y  Z  [  \\ ]  ^ _
6 ` a b c d e f g  h  i  j  k  l  m  n o
7 p q r s t u v w  x  y  z  {  |  }  ~  
```
The first 32 codes (the first two rows of the table) are special control
codes, the most common of which, `0x0a` denotes a new line (`\n`). The special
code `0x00` often denotes the end of the input, and R does not allow this
value in character strings. Code `0x7f` corresponds to a "delete" control.

When you call `utf8_print`, it uses the low level `utf8_encode` subroutine
format control codes; they format as `\uXXXX` for four hexadecimal digits
`XXXX` or as `\UXXXXYYYY` for eight hexadecimal digits `XXXXYYYY`:


```r
utf8_print(intToUtf8(1:0x0f), quote = FALSE)
```

```
[1] \u0001\u0002\u0003\u0004\u0005\u0006\a\b\t\n\v\f\r\u000e\u000f
```

Compare `utf8_print` output with the output with the base R print function:


```r
print(intToUtf8(1:0x0f), quote = FALSE)
```

```
[1] \001\002\003\004\005\006\a\b\t\n\v\f\r\016\017
```

Base R format control codes below 128 using octal escapes. There are some
other differences between the function which we will highlight below.


### Latin-1

ASCII works fine for most text in English, but not for other languages. The
Latin-1 encoding extends ASCII to Latin languages by assigning the numbers
128 to 255 (hexadecimal 0x80 to 0xff) to other common characters in Latin
languages. We can see these characters below.


```r
codes <- matrix(128:255, 8, 16, byrow = TRUE,
                dimnames = list(c(8:9, letters[1:6]), c(0:9, letters[1:6])))
latin1 <- apply(codes, c(1, 2), intToUtf8)

# replace control codes with ""
latin1[c("8", "9"),] <- ""

utf8_print(latin1, quote = FALSE)
```

```
  0 1 2 3 4 5 6 7 8 9 a b c d e f
8                                
9                                
a   ¡ ¢ £ ¤ ¥ ¦ § ¨ © ª « ¬   ® ¯
b ° ± ² ³ ´ µ ¶ · ¸ ¹ º » ¼ ½ ¾ ¿
c À Á Â Ã Ä Å Æ Ç È É Ê Ë Ì Í Î Ï
d Ð Ñ Ò Ó Ô Õ Ö × Ø Ù Ú Û Ü Ý Þ ß
e à á â ã ä å æ ç è é ê ë ì í î ï
f ð ñ ò ó ô õ ö ÷ ø ù ú û ü ý þ ÿ
```

As with ASCII, the first 32 numbers are control codes. The others are
characters common in Latin languages. Note that `0xa3`, the invalid byte
from _Mansfield Park_, corresponds to a pound sign in the Latin-1 encoding.
Given the context of the byte:


```r
lines[15252]
```

```
[1] "the command of her beauty, and her \xa320,000, any one who could satisfy the"
```

this is probably the right symbol. The text is probably encoded in Latin-1,
not UTF-8 or ASCII as claimed in the file.

If you run into an error while reading text that claims to be ASCII, it
is probably encoded as Latin-1. Note, however, that this is not the only
possibility, and there are many other encodings. The `iconvlist` function
will list the ones that R knows how to process:


```r
head(iconvlist(), n = 20)
```

```
 [1] "437"            "850"            "852"            "855"           
 [5] "857"            "860"            "861"            "862"           
 [9] "863"            "865"            "866"            "869"           
[13] "ANSI_X3.4-1968" "ANSI_X3.4-1986" "ARABIC"         "ARMSCII-8"     
[17] "ASCII"          "ASMO-708"       "ATARI"          "ATARIST"       
```

### UTF-8

With only 256 unique values, a single byte is not enough to encode every
character. Multi-byte encodings allow for encoding more. UTF-8 encodes
characters using between 1 and 4 bytes each and allows for up to 1,112,064
character codes. Most of these codes are currently unassigned, but every year
the Unicode consortium meets and adds new characters. You can find a list of
all of the characters in the [Unicode Character Database][unicode-data]. A
listing of the Emoji characters is [available separately][emoji-data].


Say you want to input the Unicode character with hexadecimal code 0x2603.
You can do so in one of three ways:

```r
"\u2603"           # with \u + 4 hex digits
```

```
[1] "☃"
```

```r
"\U00002603"       # with \U + 8 hex digits
```

```
[1] "☃"
```

```r
intToUtf8(0x2603)  # from an integer
```

```
[1] "☃"
```
For characters above `0xffff`, the first method won't work. On Windows,
a bug in the current version of R (fixed in R-devel) prevents using the
second method.


When you try to print Unicode in R, the system will first try to determine
whether the code is printable or not. Non-printable codes include control
codes and unassigned codes. On Mac OS, R uses an outdated function to make
this determination, so it is unable to print most emoji. The `utf8_print`
function uses the most recent version (10.0.0) of the Unicode standard,
and will print all Unicode characters supported by your system:


```r
print(intToUtf8(0x1f600 + 0:79)) # base R
```

```
[1] "\U0001f600\U0001f601\U0001f602\U0001f603\U0001f604\U0001f605\U0001f606\U0001f607\U0001f608\U0001f609\U0001f60a\U0001f60b\U0001f60c\U0001f60d\U0001f60e\U0001f60f\U0001f610\U0001f611\U0001f612\U0001f613\U0001f614\U0001f615\U0001f616\U0001f617\U0001f618\U0001f619\U0001f61a\U0001f61b\U0001f61c\U0001f61d\U0001f61e\U0001f61f\U0001f620\U0001f621\U0001f622\U0001f623\U0001f624\U0001f625\U0001f626\U0001f627\U0001f628\U0001f629\U0001f62a\U0001f62b\U0001f62c\U0001f62d\U0001f62e\U0001f62f\U0001f630\U0001f631\U0001f632\U0001f633\U0001f634\U0001f635\U0001f636\U0001f637\U0001f638\U0001f639\U0001f63a\U0001f63b\U0001f63c\U0001f63d\U0001f63e\U0001f63f\U0001f640\U0001f641\U0001f642\U0001f643\U0001f644\U0001f645\U0001f646\U0001f647\U0001f648\U0001f649\U0001f64a\U0001f64b\U0001f64c\U0001f64d\U0001f64e\U0001f64f"
```

```r
utf8_print(intToUtf8(0x1f600 + 0:79)) # truncates to line width
```

```
[1] "😀​😁​😂​😃​😄​😅​😆​😇​😈​😉​😊​😋​😌​😍​😎​😏​😐​😑​😒​😓​😔​😕​😖​😗​😘​😙​😚​😛​😜​😝​😞​😟​😠​😡​😢​😣​…"
```

```r
utf8_print(intToUtf8(0x1f600 + 0:79), chars = 500) # increase character limit
```

```
[1] "😀​😁​😂​😃​😄​😅​😆​😇​😈​😉​😊​😋​😌​😍​😎​😏​😐​😑​😒​😓​😔​😕​😖​😗​😘​😙​😚​😛​😜​😝​😞​😟​😠​😡​😢​😣​😤​😥​😦​😧​😨​😩​😪​😫​😬​😭​😮​😯​😰​😱​😲​😳​😴​😵​😶​😷​😸​😹​😺​😻​😼​😽​😾​😿​🙀​🙁​🙂​🙃​🙄​🙅​🙆​🙇​🙈​🙉​🙊​🙋​🙌​🙍​🙎​🙏​"
```

(Characters with codes above 0xffff, including most emoji, are not
supported on Windows.)

The *utf8* package provides the following utilities for validating, formatting,
and printing UTF-8 characters:

 + `as_utf8()` attempts to convert character data to UTF-8, throwing an
   error if the data is invalid;

 + `utf8_valid()` tests whether character data is valid according to its
   declared encoding;

 + `utf8_normalize()` converts text to Unicode composed normal form (NFC),
   optionally applying case-folding and compatibility maps;

 + `utf8_encode()` encodes a character string, escaping all control
   characters, so that it can be safely printed to the screen;

 + `utf8_format()` formats a character vector by truncating to a specified
   character width limit or by left, right, or center justifying;

 + `utf8_print()` prints UTF-8 character data to the screen;

 + `utf8_width()` measures the display with of UTF-8 character strings
    (many emoji and East Asian characters are twice as wide as other
    characters).

The package does not provide a method to translate from another encoding to
UTF-8 as the `iconv()` function from base R already serves this purpose.


Translating to UTF-8
--------------------

Back to our original problem: getting the text of _Mansfield Park_ into R.
Our first attempt failed:


```r
corpus::term_stats(lines)
```

```
Error in corpus::term_stats(lines): argument entry 15252 is incorrectly marked as "UTF-8": invalid leading byte (0xA3) at position 36
```

We discovered a problem on line 15252:


```r
lines[15252]
```

```
[1] "the command of her beauty, and her \xa320,000, any one who could satisfy the"
```

The text is likely encoded in Latin-1, not UTF-8 (or ASCII) as we had
originally thought. We can test this by attempting to convert from
Latin-1 to UTF-8 with the `iconv()` function and inspecting the output:


```r
lines2 <- iconv(lines, "latin1", "UTF-8")
lines2[15252]
```

```
[1] "the command of her beauty, and her £20,000, any one who could satisfy the"
```

It worked! Now we can analyze our text.


```r
f <- corpus::text_filter(drop_punct = TRUE, drop = corpus::stopwords_en)
corpus::term_stats(lines2, f)
```

```
   term     count support
1  fanny      816     806
2  must       508     492
3  crawford   493     488
4  mr         482     466
5  much       459     450
6  miss       432     419
7  said       406     400
8  mrs        408     399
9  sir        372     366
10 edmund     364     364
11 one        370     358
12 think      349     346
13 now        333     331
14 might      324     320
15 time       310     307
16 little     309     300
17 nothing    301     291
18 well       299     286
19 thomas     288     285
20 good       280     275
⋮  (8450 rows total)
```

The *readtext* package
----------------------

If you need more than reading in a single text file, the [readtext][readtext]
package supports reading in text in a variety of file formats and encodings.
Beyond just plain text, that package can read in PDFs, Word documents, RTF,
and many other formats. (Unfortunately, that package currently fails when
trying to read in _Mansfield Park_; the authors are aware of the issue and are
working on a fix.)


Summary
-------

Text comes in a variety of encodings, and you cannot analyze a text without
first knowing its encoding. Many functions for reading in text assume that it
is encoded in UTF-8, but this assumption sometimes fails to hold.  If you get
an error message reporting that your UTF-8 text is invalid, use `utf8_valid`
to find the offending texts. Try printing the data to the console before and
after using `iconv` to convert between character encodings. You can use
`utf8_print` to print UTF-8 characters that R refuses to display, including
emoji characters. For reading in exotic file formats like PDF or Word, try
the [readtext][readtext] package.



[emoji-data]: http://www.unicode.org/Public/emoji/5.0/emoji-data.txt

[gutenberg]: http://www.gutenberg.org

[readr]: https://github.com/tidyverse/readr#readme

[readtext]: https://github.com/quanteda/readtext

[spolsky2003]: https://www.joelonsoftware.com/2003/10/08/the-absolute-minimum-every-software-developer-absolutely-positively-must-know-about-unicode-and-character-sets-no-excuses/

[unicode]: http://unicode.org/charts/

[unicode-data]: http://www.unicode.org/Public/10.0.0/ucd/UnicodeData.txt

[utf8]: https://en.wikipedia.org/wiki/UTF-8

[windows1252]: https://en.wikipedia.org/wiki/Windows-1252

# 'format' can handle long text in Unicode locale

    Code
      raw <- c(NA, "", "a", "ab", "foo", "food", "short text", "性", "性性", "性?")
      Encoding(raw) <- "UTF-8"
      body(mapper)
    Output
      x
    Code
      mapper(utf8_format(raw, chars = 2, justify = "none", na.print = "NA"))
    Output
       [1] "NA"  ""    "a"   "ab"  "fo…" "fo…" "sh…" "性"  "性…" "性…"
    Code
      mapper(utf8_format(raw, chars = 2, justify = "left", na.print = "NA"))
    Output
       [1] "NA " "   " "a  " "ab " "fo…" "fo…" "sh…" "性 " "性…" "性…"
    Code
      mapper(utf8_format(raw, chars = 2, justify = "centre", na.print = "NA"))
    Output
       [1] "NA " "   " " a " "ab " "fo…" "fo…" "sh…" "性 " "性…" "性…"
    Code
      mapper(utf8_format(raw, chars = 2, justify = "right", na.print = "NA"))
    Output
       [1] " NA" "   " "  a" " ab" "…oo" "…od" "…xt" " 性" "…性" " …?"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{utf8_normalize}
\alias{utf8_normalize}
\title{Text Normalization}
\usage{
utf8_normalize(
  x,
  ...,
  map_case = FALSE,
  map_compat = FALSE,
  map_quote = FALSE,
  remove_ignorable = FALSE
)
}
\arguments{
\item{x}{character object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{map_case}{a logical value indicating whether to apply Unicode case
mapping to the text. For most languages, this transformation changes
uppercase characters to their lowercase equivalents.}

\item{map_compat}{a logical value indicating whether to apply Unicode
compatibility mappings to the characters, those required for NFKC and NFKD
normal forms.}

\item{map_quote}{a logical value indicating whether to replace curly single
quotes and Unicode apostrophe characters with ASCII apostrophe (U+0027).}

\item{remove_ignorable}{a logical value indicating whether to remove Unicode
"default ignorable" characters like zero-width spaces and soft hyphens.}
}
\value{
The result is a character object with the same attributes as
\code{x} but with \code{Encoding} set to \code{"UTF-8"}.
}
\description{
Transform text to normalized form, optionally mapping to lowercase and
applying compatibility maps.
}
\details{
\code{utf8_normalize()} converts the elements of a character object to Unicode
normalized composed form (NFC) while applying the character maps specified
by the \code{map_case}, \code{map_compat}, \code{map_quote}, and
\code{remove_ignorable} arguments.
}
\examples{

angstrom <- c("\u00c5", "\u0041\u030a", "\u212b")
utf8_normalize(angstrom) == "\u00c5"

}
\seealso{
\code{\link[=as_utf8]{as_utf8()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8_encode.R
\name{utf8_encode}
\alias{utf8_encode}
\title{Encode Character Object as for UTF-8 Printing}
\usage{
utf8_encode(
  x,
  ...,
  width = 0L,
  quote = FALSE,
  justify = "left",
  escapes = NULL,
  display = FALSE,
  utf8 = NULL
)
}
\arguments{
\item{x}{character object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{width}{integer giving the minimum field width; specify \code{NULL} or
\code{NA} for no minimum.}

\item{quote}{logical scalar indicating whether to surround results with
double-quotes and escape internal double-quotes.}

\item{justify}{justification; one of \code{"left"}, \code{"right"},
\code{"centre"}, or \code{"none"}. Can be abbreviated.}

\item{escapes}{a character string specifying the display style for the
backslash escapes, as an ANSI SGR parameter string, or NULL for no styling.}

\item{display}{logical scalar indicating whether to optimize the encoding
for display, not byte-for-byte data transmission.}

\item{utf8}{logical scalar indicating whether to encode for a UTF-8 capable
display (ASCII-only otherwise), or \code{NULL} to encode for output
capabilities as determined by \code{output_utf8()}.}
}
\value{
A character object with the same attributes as \code{x} but with
\code{Encoding} set to \code{"UTF-8"}.
}
\description{
Escape the strings in a character object, optionally adding quotes or
spaces, adjusting the width for display.
}
\details{
\code{utf8_encode()} encodes a character object for printing on a UTF-8 device
by escaping controls characters and other non-printable characters. When
\code{display = TRUE}, the function optimizes the encoding for display by
removing default ignorable characters (soft hyphens, zero-width spaces,
etc.) and placing zero-width spaces after wide emoji. When
\code{output_utf8()} is \code{FALSE} the function escapes all non-ASCII
characters and gives the same results on all platforms.
}
\examples{

# the second element is encoded in latin-1, but declared as UTF-8
x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
Encoding(x) <- c("UTF-8", "UTF-8", "bytes")

# encoding
utf8_encode(x)

# add style to the escapes
cat(utf8_encode("hello\nstyled\\\\world", escapes = "1"), "\n")

}
\seealso{
\code{\link[=utf8_print]{utf8_print()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/output.R
\name{output_ansi}
\alias{output_ansi}
\alias{output_utf8}
\title{Output Capabilities}
\usage{
output_ansi()

output_utf8()
}
\value{
A logical scalar indicating whether the output connection supports
the given capability.
}
\description{
Test whether the output connection has ANSI style escape support or UTF-8
support.
}
\details{
\code{output_ansi()} tests whether the output connection supports ANSI style
escapes. This is \code{TRUE} if the connection is a terminal and not the
Windows GUI. Otherwise, it is true if running in RStudio 1.1 or later with
ANSI escapes enabled, provided \code{stdout()} has not been redirected to
another connection by \code{sink()}.

\code{output_utf8()} tests whether the output connection supports UTF-8. For
most platforms \code{l10n_info()$"UTF-8"} gives this information, but this
does not give an accurate result for Windows GUIs. To work around this, we
proceed as follows: \itemize{

\item if the character locale
(\code{LC_CTYPE}) is \code{"C"}, then the result is \code{FALSE};

\item otherwise, if \code{l10n_info()$"UTF-8"} is \code{TRUE}, then the
result is \code{TRUE};

\item if running on Windows, then the result is \code{TRUE};

\item in all other cases the result is \code{FALSE}. }

Strictly speaking,
UTF-8 support is always available on Windows GUI, but only a subset of UTF-8
is available (defined by the current character locale) when the output is
redirected by \code{knitr} or another process. Unfortunately, it is
impossible to set the character locale to UTF-8 on Windows. Further, the
\code{utf8} package only handles two character locales: C and UTF-8.  To get
around this, on Windows, we treat all non-C locales on that platform as
UTF-8. This liberal approach means that characters in the user's locale
never get escaped; others will get output as \verb{<U+XXXX>}, with incorrect
values for \code{utf8_width()}.
}
\examples{

# test whether ANSI style escapes or UTF-8 output are supported
cat("ANSI:", output_ansi(), "\n")
cat("UTF8:", output_utf8(), "\n")

# switch to C locale
Sys.setlocale("LC_CTYPE", "C")
cat("ANSI:", output_ansi(), "\n")
cat("UTF8:", output_utf8(), "\n")

# switch to native locale
Sys.setlocale("LC_CTYPE", "")

tmp <- tempfile()
sink(tmp) # redirect output to a file
cat("ANSI:", output_ansi(), "\n")
cat("UTF8:", output_utf8(), "\n")
sink() # restore stdout

# inspect the output
readLines(tmp)

}
\seealso{
\code{\link[=.Platform]{.Platform()}}, \code{\link[=isatty]{isatty()}},
\code{\link[=l10n_info]{l10n_info()}}, \code{\link[=Sys.getlocale]{Sys.getlocale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{utf8_width}
\alias{utf8_width}
\title{Measure the Character String Width}
\usage{
utf8_width(x, ..., encode = TRUE, quote = FALSE, utf8 = NULL)
}
\arguments{
\item{x}{character object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{encode}{whether to encode the object before measuring its width.}

\item{quote}{whether to quote the object before measuring its width.}

\item{utf8}{logical scalar indicating whether to determine widths assuming a
UTF-8 capable display (ASCII-only otherwise), or \code{NULL} to format for
output capabilities as determined by \code{output_utf8()}.}
}
\value{
An integer object, with the same \code{names}, \code{dim}, and
\code{dimnames} as \code{x}.
}
\description{
Compute the display widths of the elements of a character object.
}
\details{
\code{utf8_width()} returns the printed widths of the elements of a character
object on a UTF-8 device (or on an ASCII device when \code{output_utf8()} is
\code{FALSE}), when printed with \code{utf8_print()}. If the string is not
printable on the device, for example if it contains a control code like
\code{"\\n"}, then the result is \code{NA}. If \code{encode = TRUE}, the
default, then the function returns the widths of the encoded elements via
\code{utf8_encode()}); otherwise, the function returns the widths of the
original elements.
}
\examples{

# the second element is encoded in latin-1, but declared as UTF-8
x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
Encoding(x) <- c("UTF-8", "UTF-8", "bytes")

# get widths
utf8_width(x)
utf8_width(x, encode = FALSE)
utf8_width('"')
utf8_width('"', quote = TRUE)

}
\seealso{
\code{\link[=utf8_print]{utf8_print()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8_print.R
\name{utf8_print}
\alias{utf8_print}
\title{Print UTF-8 Text}
\usage{
utf8_print(
  x,
  ...,
  chars = NULL,
  quote = TRUE,
  na.print = NULL,
  print.gap = NULL,
  right = FALSE,
  max = NULL,
  names = NULL,
  rownames = NULL,
  escapes = NULL,
  display = TRUE,
  style = TRUE,
  utf8 = NULL
)
}
\arguments{
\item{x}{character object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{chars}{integer scalar indicating the maximum number of character units
to display.  Wide characters like emoji take two character units; combining
marks and default ignorables take none. Longer strings get truncated and
suffixed or prefixed with an ellipsis (\code{"..."} in C locale,
\code{"\\u2026"} in others). Set to \code{NULL} to limit output to the line
width as determined by \code{getOption("width")}.}

\item{quote}{logical scalar indicating whether to put surrounding
double-quotes (\code{'"'}) around character strings and escape internal
double-quotes.}

\item{na.print}{character string (or \code{NULL}) indicating the encoding
for \code{NA} values. Ignored when \code{na.encode} is \code{FALSE}.}

\item{print.gap}{non-negative integer (or \code{NULL}) giving the number of
spaces in gaps between columns; set to \code{NULL} or \code{1} for a single
space.}

\item{right}{logical scalar indicating whether to right-justify character
strings.}

\item{max}{non-negative integer (or \code{NULL}) indicating the maximum
number of elements to print; set to \code{getOption("max.print")} if
argument is \code{NULL}.}

\item{names}{a character string specifying the display style for the
(column) names, as an ANSI SGR parameter string.}

\item{rownames}{a character string specifying the display style for the row
names, as an ANSI SGR parameter string.}

\item{escapes}{a character string specifying the display style for the
backslash escapes, as an ANSI SGR parameter string.}

\item{display}{logical scalar indicating whether to optimize the encoding
for display, not byte-for-byte data transmission.}

\item{style}{logical scalar indicating whether to apply ANSI terminal escape
codes to style the output.  Ignored when \code{output_ansi()} is
\code{FALSE}.}

\item{utf8}{logical scalar indicating whether to optimize results for a
UTF-8 capable display, or \code{NULL} to set as the result of
\code{output_utf8()}. Ignored when \code{output_utf8()} is \code{FALSE}.}
}
\value{
The function returns \code{x} invisibly.
}
\description{
Print a UTF-8 character object.
}
\details{
\code{utf8_print()} prints a character object after formatting it with
\code{\link[=utf8_format]{utf8_format()}}.

For ANSI terminal output (when \code{output_ansi()} is \code{TRUE}), you can
style the row and column names with the \code{rownames} and \code{names}
parameters, specifying an ANSI SGR parameter string; see
\url{https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters}.
}
\examples{

# printing (assumes that output is capable of displaying Unicode 10.0.0)
print(intToUtf8(0x1F600 + 0:79)) # with default R print function
utf8_print(intToUtf8(0x1F600 + 0:79)) # with utf8_print, truncates line
utf8_print(intToUtf8(0x1F600 + 0:79), chars = 1000) # higher character limit

# in C locale, output ASCII (same results on all platforms)
oldlocale <- Sys.getlocale("LC_CTYPE")
invisible(Sys.setlocale("LC_CTYPE", "C")) # switch to C locale
utf8_print(intToUtf8(0x1F600 + 0:79))
invisible(Sys.setlocale("LC_CTYPE", oldlocale)) # switch back to old locale

# Mac and Linux only: style the names
# see https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
utf8_print(matrix(as.character(1:20), 4, 5),
           names = "1;4", rownames = "2;3")

}
\seealso{
\code{\link[=utf8_format]{utf8_format()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{as_utf8}
\alias{as_utf8}
\alias{utf8_valid}
\title{UTF-8 Character Encoding}
\usage{
as_utf8(x, normalize = FALSE)

utf8_valid(x)
}
\arguments{
\item{x}{character object.}

\item{normalize}{a logical value indicating whether to convert to Unicode
composed normal form (NFC).}
}
\value{
For \code{as_utf8()}, the result is a character object with the same
attributes as \code{x} but with \code{Encoding} set to \code{"UTF-8"}.

For \code{utf8_valid()} a logical object with the same \code{names},
\code{dim}, and \code{dimnames} as \code{x}.
}
\description{
UTF-8 text encoding and validation

\code{as_utf8()} converts a character object from its declared encoding to a
valid UTF-8 character object, or throws an error if no conversion is
possible. If \code{normalize = TRUE}, then the text gets transformed to
Unicode composed normal form (NFC) after conversion to UTF-8.

\code{utf8_valid()} tests whether the elements of a character object can be
translated to valid UTF-8 strings.
}
\examples{

# the second element is encoded in latin-1, but declared as UTF-8
x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
Encoding(x) <- c("UTF-8", "UTF-8", "bytes")

# attempt to convert to UTF-8 (fails)
\dontrun{as_utf8(x)}

y <- x
Encoding(y[2]) <- "latin1" # mark the correct encoding
as_utf8(y) # succeeds

# test for valid UTF-8
utf8_valid(x)

}
\seealso{
\code{\link[=utf8_normalize]{utf8_normalize()}}, \code{\link[=iconv]{iconv()}}.
}

\name{output_utf8}
\title{Output Capabilities}
\alias{output_ansi}
\alias{output_utf8}
\description{
    Test whether the output connection has ANSI style escape support
    or UTF-8 support.
}
\usage{
output_ansi()

output_utf8()
}
\details{
\code{output_ansi} tests whether the output connection supports ANSI
style escapes. This is \code{TRUE} if the connection is a terminal
and not the Windows GUI. Otherwise, it is true if running in
RStudio 1.1 or later with ANSI escapes enabled, provided \code{stdout()}
has not been redirected to another connection by \code{sink()}.

\code{output_utf8} tests whether the output connection supports UTF-8. For
most platforms \code{l10n_info()$`UTF-8`} gives this information, but this
does not give an accurate result for Windows GUIs. To work around this,
we proceed as follows:
\itemize{
  \item if the character locale (\code{LC_CTYPE}) is \code{"C"}, then
      the result is \code{FALSE};

  \item otherwise, if \code{l10n_info()$`UTF-8`} is \code{TRUE},
      then the result is \code{TRUE};

  \item if running on Windows, then the result is \code{TRUE};
    
  \item in all other cases the result is \code{FALSE}.
}
Strictly speaking, UTF-8 support is always available on Windows GUI,
but only a subset of UTF-8 is available (defined by the current character
locale) when the output is redirected by \code{knitr} or another process.
Unfortunately, it is impossible to set the character locale to UTF-8 on
Windows. Further, the \code{utf8} package only handles two character
locales: C and UTF-8.  To get around this, on Windows, we treat all
non-C locales on that platform as UTF-8. This liberal approach means that
characters in the user's locale never get escaped; others will get output
as \code{<U+XXXX>}, with incorrect values for \code{utf8_width}.
}
\value{
A logical scalar indicating whether the output connection supports the
given capability.
}
\seealso{
    \code{\link{.Platform}}, \code{\link{isatty}}, \code{\link{l10n_info}},
    \code{\link{Sys.getlocale}}
}
\examples{
# test whether ANSI style escapes or UTF-8 output are supported
cat("ANSI:", output_ansi(), "\n")
cat("UTF8:", output_utf8(), "\n")

# switch to C locale
Sys.setlocale("LC_CTYPE", "C")
cat("ANSI:", output_ansi(), "\n")
cat("UTF8:", output_utf8(), "\n")

# switch to native locale
Sys.setlocale("LC_CTYPE", "")

tmp <- tempfile()
sink(tmp) # redirect output to a file
cat("ANSI:", output_ansi(), "\n")
cat("UTF8:", output_utf8(), "\n")
sink() # restore stdout

# inspect the output
readLines(tmp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8_format.R
\name{utf8_format}
\alias{utf8_format}
\title{UTF-8 Text Formatting}
\usage{
utf8_format(
  x,
  ...,
  trim = FALSE,
  chars = NULL,
  justify = "left",
  width = NULL,
  na.encode = TRUE,
  quote = FALSE,
  na.print = NULL,
  print.gap = NULL,
  utf8 = NULL
)
}
\arguments{
\item{x}{character object.}

\item{...}{These dots are for future extensions and must be empty.}

\item{trim}{logical scalar indicating whether to suppress padding spaces
around elements.}

\item{chars}{integer scalar indicating the maximum number of character units
to display.  Wide characters like emoji take two character units; combining
marks and default ignorables take none. Longer strings get truncated and
suffixed or prefixed with an ellipsis (\code{"..."} or \code{"\\u2026"},
whichever is most appropriate for the current character locale). Set to
\code{NULL} to limit output to the line width as determined by
\code{getOption("width")}.}

\item{justify}{justification; one of \code{"left"}, \code{"right"},
\code{"centre"}, or \code{"none"}. Can be abbreviated.}

\item{width}{the minimum field width; set to \code{NULL} or \code{0} for no
restriction.}

\item{na.encode}{logical scalar indicating whether to encode \code{NA}
values as character strings.}

\item{quote}{logical scalar indicating whether to format for a context with
surrounding double-quotes (\code{'"'}) and escaped internal double-quotes.}

\item{na.print}{character string (or \code{NULL}) indicating the encoding
for \code{NA} values. Ignored when \code{na.encode} is \code{FALSE}.}

\item{print.gap}{non-negative integer (or \code{NULL}) giving the number of
spaces in gaps between columns; set to \code{NULL} or \code{1} for a single
space.}

\item{utf8}{logical scalar indicating whether to format for a UTF-8 capable
display (ASCII-only otherwise), or \code{NULL} to format for output
capabilities as determined by \code{output_utf8()}.}
}
\value{
A character object with the same attributes as \code{x} but with
\code{Encoding} set to \code{"UTF-8"} for elements that can be converted to
valid UTF-8 and \code{"bytes"} for others.
}
\description{
Format a character object for UTF-8 printing.
}
\details{
\code{utf8_format()} formats a character object for printing, optionally
truncating long character strings.
}
\examples{

# the second element is encoded in latin-1, but declared as UTF-8
x <- c("fa\u00E7ile", "fa\xE7ile", "fa\xC3\xA7ile")
Encoding(x) <- c("UTF-8", "UTF-8", "bytes")

# formatting
utf8_format(x, chars = 3)
utf8_format(x, chars = 3, justify = "centre", width = 10)
utf8_format(x, chars = 3, justify = "right")

}
\seealso{
\code{\link[=utf8_print]{utf8_print()}}, \code{\link[=utf8_encode]{utf8_encode()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8-package.R
\docType{package}
\name{utf8-package}
\alias{utf8}
\alias{utf8-package}
\title{The utf8 Package}
\description{
UTF-8 Text Processing
}
\details{
Functions for manipulating and printing UTF-8 encoded text:

\itemize{ \item \code{\link[=as_utf8]{as_utf8()}} attempts to convert character data to
UTF-8, throwing an error if the data is invalid;

\item \code{\link[=utf8_valid]{utf8_valid()}} tests whether character data is valid
according to its declared encoding;

\item \code{\link[=utf8_normalize]{utf8_normalize()}} converts text to Unicode composed normal
form (NFC), optionally applying case-folding and compatibility maps;

\item \code{\link[=utf8_encode]{utf8_encode()}} encodes a character string, escaping all
control characters, so that it can be safely printed to the screen;

\item \code{\link[=utf8_format]{utf8_format()}} formats a character vector by truncating to
a specified character width limit or by left, right, or center justifying;

\item \code{\link[=utf8_print]{utf8_print()}} prints UTF-8 character data to the screen;

\item \code{\link[=utf8_width]{utf8_width()}} measures the display width of UTF-8 character
strings (many emoji and East Asian characters are twice as wide as other
characters);

\item \code{\link[=output_ansi]{output_ansi()}} and \code{\link[=output_utf8]{output_utf8()}} test for the
output connections capabilities. }

For a complete list of functions, use \code{library(help = "utf8")}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://krlmlr.github.io/utf8/}
  \item \url{https://github.com/krlmlr/utf8}
  \item Report bugs at \url{https://github.com/krlmlr/utf8/issues}
}

}
\author{
\strong{Maintainer}: Kirill Müller \email{kirill@cynkra.com} (\href{https://orcid.org/0000-0002-1416-3412}{ORCID})

Authors:
\itemize{
  \item Patrick O. Perry [copyright holder]
}

Other contributors:
\itemize{
  \item Unicode, Inc. (Unicode Character Database) [copyright holder, data contributor]
}

}
